import pandas as pd

def filter_empty_genes(counts: pd.DataFrame) -> pd.DataFrame:
    """
    Remove gene with all counts values to zero
    """
    if counts.empty:
        return counts
    
    select_columns = counts.apply(lambda col: col.sum() > 0)
    select_columns = counts.columns[select_columns]
    filtered_counts = counts[select_columns]

    return filtered_counts


import scanpy as sc
import numpy as np
import pandas as pd
from .build_reference import rank_preprocess, get_fastccc_input
from .build_reference import calculate_L_R_and_IS_score, calculate_L_R_and_IS_percents
from .build_reference import calculate_mean_pmfs, record_hk_genes
from . import score
from .core import calculate_cluster_percents
from loguru import logger
import pickle
import itertools
from scipy.stats import norm
import os
import tomllib
from collections import Counter
from .distrib_digit import get_minimum_distribution_for_digit
from . import dist_complex
from . import dist_lr
import json

# logger.remove()  # 移除默认的日志处理器
# logger.add(
#     sink=lambda msg: print(msg.strip()),
#     format="<green>{time:YYYY-MM-DD HH:mm:ss}</green> | <level>{level}</level> | <cyan>{message}</cyan>",
#     colorize=True
# )



precision = 0.01

def get_norm_minimum_ppf_func(locs, scales):
    def analytic_func(x, locs=locs, scales=scales):
        F = []
        for loc, scale in zip(locs, scales):
            F.append(norm.cdf(x, loc, scale))
        minF = 1 - np.prod(1-np.array(F), axis=0)
        return minF
    def inverse_f(y, low=0, high=50, tol=1e-6): # high=n_bins
        assert y > tol and y < 1 - tol
        while high - low > tol:
            mid = (low + high) / 2
            if analytic_func(mid) < y:
                low = mid
            else:
                high = mid
        return (low + high) / 2
    return inverse_f

def get_norm_ppf_func(loc, scale):
    def analytic_func(x):
        return norm.ppf(x, loc, scale)
    return analytic_func
    
def get_threshold_from_pmf(pmf, alpha=0.025):
    '''
    alpha: 概率分布右侧面积
    '''
    if pmf.is_analytic:
        threshold = get_norm_ppf_func(pmf.loc, pmf.scale)(1-alpha)
    elif pmf.is_complex_analytic:
        raise RuntimeError
    else:
        y = pmf.get_cdf_array()
        threshold = np.argmax((1 - y) <= alpha) * precision
    return threshold

def compare_strategy():
    pass

def get_valid_pval_pmfs(mean_pmfs, interactions_strength, interactions, percents_analysis):
    p1_index = []
    p2_index = []
    all_index = []
    for i in interactions_strength.index:
        p1_index.append(i.split('|')[0])
        p2_index.append(i.split('|')[1])
        all_index.append(i)
        
    p1 = mean_pmfs.loc[p1_index, interactions['multidata_1_id']]
    p2 = mean_pmfs.loc[p2_index, interactions['multidata_2_id']]
    p1.columns = interactions.index
    p2.columns = interactions.index
    p1.index = all_index
    p2.index = all_index

    p1_items = p1.values[np.where(percents_analysis)]
    p2_items = p2.values[np.where(percents_analysis)]

    pval_pmfs = (p1_items & p2_items) / 2

    return pval_pmfs

def get_ligand_receptor_summary(mean_counts, interactions):
    p1_index = []
    p2_index = []
    all_index = []
    for i in itertools.product(sorted(mean_counts.index), sorted(mean_counts.index)):
        p1_index.append(i[0])
        p2_index.append(i[1])
        all_index.append('|'.join(i))
    p1 = mean_counts.loc[p1_index, interactions['multidata_1_id']]
    p2 = mean_counts.loc[p2_index, interactions['multidata_2_id']]
    p1.columns = interactions.index
    p2.columns = interactions.index
    p1.index = all_index
    p2.index = all_index
    interactions_strength = (p1 + p2)/2 * (p1 > 0) * (p2>0)
    return p1, p2, interactions_strength

def get_celltype_mapping_dict(celltype_mapping_dict):
    match celltype_mapping_dict:
        case None:
            logger.info("Reference cell types label will be used directly.")
            return None
        case str():
            logger.info("Parsing celltype_mapping_dict file.")
            try:
                with open(celltype_mapping_dict, 'r', encoding='utf-8') as file:
                    celltype_mapping_dict = json.load(file)
                return celltype_mapping_dict
            except FileNotFoundError:
                logger.error(f"The file '{celltype_mapping_dict}' was not found.")
                logger.info("Reference cell types label will be used directly.")
                return None
            except json.JSONDecodeError:
                logger.error(f"The file '{celltype_mapping_dict}' is not a valid JSON file.")
                logger.info("Reference cell types label will be used directly.")
                return None
            return "The variable is a string."
        case dict():
            logger.info("The variable celltype_mapping_dict is used.")
            return celltype_mapping_dict
        case _:
            logger.error(f"The variable celltype_mapping_dict is of an unknown type.")
            logger.info("Reference cell types label will be used directly.")
            return None

def aggregate_by_weight(df, merge_dict, weight_dict):
    results = []
    index = []
    for key in sorted(merge_dict):
        index.append(key)
        row = []
        sum_ = 0
        if len(merge_dict[key]) == 1:
            results.append(df.loc[merge_dict[key][0]].values)
        else:
            for ct in merge_dict[key]:
                assert weight_dict[ct] > 0
                sum_ += weight_dict[ct]
                row.append(df.loc[ct].values * weight_dict[ct])
            results.append(np.array(row).sum(axis=0) / sum_)
    return pd.DataFrame(results, index=index, columns=df.columns)


def load_LRI_id_to_gene_symbol_dict(LRI_db_path):
    from . import preprocess
    interactions = preprocess.get_interactions(LRI_db_path)
    id2symbol_dict = (pd.read_csv(f'{LRI_db_path}/gene_table.csv')[['protein_id', 'hgnc_symbol']]\
    .set_index('protein_id')['hgnc_symbol']).to_dict()
    ##### complex_table ######
    complex_composition = pd.read_csv(os.path.join(LRI_db_path, 'complex_composition_table.csv'))
    complex_table = pd.read_csv(os.path.join(LRI_db_path, 'complex_table.csv'))
    complex_table = complex_table.merge(complex_composition, left_on='complex_multidata_id', right_on='complex_multidata_id')
    foo_dict = complex_table.groupby('complex_multidata_id').apply(lambda x: list(x['protein_multidata_id'].values), include_groups=False).to_dict()
    for key in foo_dict:
        value = ','.join([id2symbol_dict[item] for item in foo_dict[key]])
        id2symbol_dict[key] = value
    return id2symbol_dict


def get_null_ref_mean_counts(ref_gene_pmf_dict, ref_complex_table):
    # 使用 reference 数据，在没有任何条件下，gene 的均值应该是多少，complex 也同样被考虑。
    # 返回一个 dataframe
    null_ref_mean_counts = []
    columns = []
    for multi_id in ref_gene_pmf_dict:
        columns.append(multi_id)
        null_ref_mean_counts.append(ref_gene_pmf_dict[multi_id][1].mean)
    null_ref_mean_counts = pd.DataFrame([null_ref_mean_counts], index=['ref_null'], columns=columns)
    complex_func = score.calculate_complex_min_func
    null_ref_mean_counts = score.combine_complex_distribution_df(null_ref_mean_counts, ref_complex_table, complex_func)
    return null_ref_mean_counts


def compare_with_reference(counts_df, labels_df, complex_table, interactions, reference_path, save_path, config, celltype_mapping_dict, database_file_path, k=2.59, debug_mode=False):
    assert os.path.exists(reference_path), "Reference dir doesn't exist."

    logger.info("Loading reference data.")
    with open(f'{reference_path}/complex_table.pkl', 'rb') as f:
        ref_complex_table = pickle.load(f)
    with open(f'{reference_path}/interactions.pkl', 'rb') as f:
        ref_interactions = pickle.load(f)
    with open(f'{reference_path}/ref_gene_pmf_dict.pkl', 'rb') as f:
        ref_gene_pmf_dict = pickle.load(f)

    with open(f'{reference_path}/ref_percents.pkl', 'rb') as f:
        ref_percents = pickle.load(f)
    with open(f'{reference_path}/ref_mean_counts.pkl', 'rb') as f:
        ref_mean_counts = pickle.load(f)

    gene_list = [gene for gene in ref_mean_counts.columns if gene in counts_df.columns]
    ref_complex_table = ref_complex_table.loc[[item for item in ref_complex_table.index if item in complex_table.index]]
    ref_interactions = ref_interactions.loc[[item for item in ref_interactions.index if item in interactions.index]]

    null_ref_mean_counts = get_null_ref_mean_counts(ref_gene_pmf_dict, ref_complex_table)

    ref_label_counter = config['celltype']
    label_counter = Counter(labels_df['cell_type'])

    # 25-01-08 add
    # Combine cell types
    celltype_mapping_dict = get_celltype_mapping_dict(celltype_mapping_dict)
    if celltype_mapping_dict is None:

        ref_meta_dict = {}
        for item in ref_label_counter:
            if item not in label_counter:
                continue
            ref_meta_dict[item] = ref_label_counter[item]
        logger.debug(f"Valid reference cell type data:\n{str(ref_meta_dict)}")

    else:
        for key, value in celltype_mapping_dict.items():
            assert key in ref_label_counter, 'Please check the spelling, capitalization, spacing, and format of “your cell type”, as well as whether it is included in the reference.'
            assert value in label_counter, 'Please check the spelling, capitalization, spacing, and format of “your cell type”, as well as whether it is included in the query.'
        
        query_to_reference_dict = {}
        for key, value in celltype_mapping_dict.items():
            if value not in query_to_reference_dict:
                query_to_reference_dict[value] = [key]
            else:
                query_to_reference_dict[value].append(key)

        for item in ref_label_counter:
            if item not in celltype_mapping_dict and item in label_counter:
                query_to_reference_dict[item] = [item]

        ref_mean_counts = aggregate_by_weight(ref_mean_counts, query_to_reference_dict, ref_label_counter)
        ref_percents = aggregate_by_weight(ref_percents, query_to_reference_dict, ref_label_counter)

        ref_meta_dict = {}
        for key, values in query_to_reference_dict.items():
            sum_ = 0
            for item in values:
                sum_ += ref_label_counter[item]
            ref_meta_dict[key] = sum_
        logger.debug(f"Valid reference cell type data:\n{str(ref_meta_dict)}")

    unique_meta_dict = {}
    for item in label_counter:
        if item not in ref_meta_dict:
            unique_meta_dict[item] = label_counter[item]
    logger.debug(f"Unique query cell type data:\n{str(unique_meta_dict)}")
    

    # 25-01-08 end


    
    # ref_meta_dict = {}
    # for item in ref_label_counter:
    #     if item not in label_counter:
    #         continue
    #     ref_meta_dict[item] = ref_label_counter[item]

    ref_mean_counts = ref_mean_counts.loc[[item for item in ref_mean_counts.index if item in ref_meta_dict]]
    ref_percents = ref_percents.loc[[item for item in ref_mean_counts.index if item in ref_meta_dict]]

    ref_p1, ref_p2, ref_CS = get_ligand_receptor_summary(ref_mean_counts, ref_interactions)
    ref_L_perc, ref_R_perc, ref_percents_analysis = calculate_L_R_and_IS_percents(ref_percents, ref_interactions, threshold=config['min_percentile'])

    ref_clusters_mean_dict = {}
    for celltype in sorted(ref_meta_dict):
        ref_clusters_mean_dict[celltype]  = {}
        n_sum = ref_meta_dict[celltype]
        if n_sum < 100:
            for gene in gene_list:
                ref_clusters_mean_dict[celltype][gene] = ref_gene_pmf_dict[gene][n_sum]
        else:
            for gene in gene_list:
                ref_clusters_mean_dict[celltype][gene] = ref_gene_pmf_dict[gene][1] ** n_sum / n_sum
    ref_mean_pmfs = pd.DataFrame(ref_clusters_mean_dict).T
    complex_func = get_minimum_distribution_for_digit
    if len(ref_mean_pmfs):
        ref_mean_pmfs = dist_complex.combine_complex_distribution_df(ref_mean_pmfs, ref_complex_table, complex_func)
    logger.success("Reference data is loaded.")

    logger.info("Calculating CS score for query data.")
    mean_counts = score.calculate_cluster_mean(counts_df, labels_df)
    null_mean_counts = mean_counts.copy()
    null_mean_counts.values[:,:] = np.repeat(np.array([counts_df.mean(axis=0)]), len(mean_counts), axis=0).reshape(*mean_counts.shape)
    complex_func = score.calculate_complex_min_func
    mean_counts = score.combine_complex_distribution_df(mean_counts, complex_table, complex_func)
    null_mean_counts = score.combine_complex_distribution_df(null_mean_counts, complex_table, complex_func)
    null_interactions_strength = score.calculate_interactions_strength(null_mean_counts, interactions, method='Arithmetic')
    p1, p2, interactions_strength = calculate_L_R_and_IS_score(mean_counts, interactions)
    percents = calculate_cluster_percents(counts_df, labels_df, complex_table)

    logger.info("Filtering reference data.")
    common_ind = sorted(set(ref_percents_analysis.index) & set(interactions_strength.index))
    common_col = sorted(set(ref_percents_analysis.columns) & set(interactions_strength.columns))
    if len(ref_mean_pmfs):
        ref_pvals = dist_lr.calculate_key_interactions_pvalue(
            ref_mean_pmfs, ref_interactions, ref_CS, ref_percents_analysis, method='Arithmetic'
        )
        ref_pvals = ref_pvals.loc[common_ind, common_col]
    else:
        ref_pvals = pd.DataFrame([], index=common_ind, columns=common_col)
    # real_ref_pvals = pd.read_csv(f'{reference_path}/ref_pvals.txt', sep='\t', index_col=0)
    # real_ref_pvals = real_ref_pvals.loc[common_ind, common_col]
    # print(f"Error: {np.sum(np.abs(real_ref_pvals.values - ref_pvals.values))}")
    # print(f"Equal: {np.array_equal(real_ref_pvals.values, ref_pvals.values)}")
    # tmp = np.where(real_ref_pvals.values!= ref_pvals.values)
    # print('real')
    # print(real_ref_pvals.values[tmp])
    # print('ref')
    # print(ref_pvals.values[tmp])
    # print(f'ref_pvals shape: {ref_pvals.shape}')
    ref_percents_analysis= ref_percents_analysis.loc[common_ind, common_col]
    ref_p1= ref_p1.loc[common_ind, common_col]
    ref_p2= ref_p2.loc[common_ind, common_col]
    ref_L_perc = ref_L_perc.loc[common_ind, common_col]
    ref_R_perc = ref_R_perc.loc[common_ind, common_col]

    

    logger.info("Filtering by using reference.")
    # filtering use ref
    complex_table = complex_table.loc[[item for item in complex_table.index if item in ref_complex_table.index]]
    interactions = interactions.loc[[item for item in interactions.index if item in ref_interactions.index]]
    L_perc, R_perc, percents_analysis = calculate_L_R_and_IS_percents(percents, interactions, threshold=config['min_percentile'])
    interactions_strength = interactions_strength.loc[percents_analysis.index, percents_analysis.columns]
    null_interactions_strength = null_interactions_strength.loc[percents_analysis.index, percents_analysis.columns]
    p1 = p1.loc[percents_analysis.index, percents_analysis.columns]
    p2 = p2.loc[percents_analysis.index, percents_analysis.columns]

    p1_list = p1.values[np.where(percents_analysis)]
    p2_list = p2.values[np.where(percents_analysis)]

    logger.info("Inferring sig. boundaries.")
    mean_pmfs = calculate_mean_pmfs(counts_df, labels_df, complex_table, ref_gene_pmf_dict)
    pval_pmfs = get_valid_pval_pmfs(mean_pmfs, interactions_strength, interactions, percents_analysis)

    up_bound_list = []
    low_bound_list = []
    for i in range(len(pval_pmfs)):
        threshold = get_threshold_from_pmf(pval_pmfs[i], 0.05)
        up_bound = threshold + threshold / k * 1.96 
        low_bound = threshold - threshold / k * 1.96 
        up_bound_list.append(up_bound)
        low_bound_list.append(low_bound)

    valid_IS_list = interactions_strength.values[np.where(percents_analysis)]
    null_IS_list = null_interactions_strength.values[np.where(percents_analysis)]
    L_perc_list = L_perc.values[np.where(percents_analysis)]
    R_perc_list = R_perc.values[np.where(percents_analysis)]

    id2symbol_dict = load_LRI_id_to_gene_symbol_dict(database_file_path)

    # strategy
    prediction = []
    est_by_ref = []
    results = []
    for i, (index, col) in enumerate(zip(*np.where(percents_analysis))):
        index = percents_analysis.index[index]
        col = percents_analysis.columns[col]

        ligand_gene_name = interactions.loc[col].multidata_1_id
        receptor_gene_name = interactions.loc[col].multidata_2_id
        ligand_gene_name = id2symbol_dict[ligand_gene_name]
        receptor_gene_name =  id2symbol_dict[receptor_gene_name]

        #01-11 add
        if debug_mode == True:
            mid1 = interactions.loc[col].multidata_1_id
            mid2 = interactions.loc[col].multidata_2_id
            # L_null_ref = ref_mean_pmfs.iloc[0][mid1].mean
            # R_null_ref = ref_mean_pmfs.iloc[0][mid2].mean
            L_null_ref = null_ref_mean_counts[mid1].item()
            R_null_ref = null_ref_mean_counts[mid2].item()

        #01-11 end

        IS = interactions_strength.loc[index, col]
        assert valid_IS_list[i] == IS
        low_IS = low_bound_list[i]
        up_IS = up_bound_list[i]
        null_IS = null_IS_list[i]

        ligand_IS = p1_list[i]
        receptor_IS = p2_list[i]
        ligand_perc = L_perc_list[i]
        receptor_perc = R_perc_list[i]

        ref_flag = index in ref_pvals.index
        is_in_ref = True if ref_flag else False
        ref_perc = ref_percents_analysis.loc[index, col] if ref_flag else np.nan
        ref_ligand_perc = ref_L_perc.loc[index, col] if ref_flag else np.nan
        ref_receptor_perc = ref_R_perc.loc[index, col] if ref_flag else np.nan
        ligand_low = np.max(ref_p1.loc[index, col] * (1-1/k*1.96), 0) if ref_flag else np.nan
        ligand_high = ref_p1.loc[index, col] * (1+1/k*1.96) if ref_flag else np.nan
        receptor_low = np.max(ref_p2.loc[index, col], 0) * (1-1/k*1.96) if ref_flag else np.nan
        receptor_high = ref_p2.loc[index, col] * (1+1/k*1.96) if ref_flag else np.nan
        ref_sig = (ref_pvals.loc[index, col] < 0.05) if ref_flag else np.nan # 参考是否显著
        ligand_range = f"{ligand_low}-{ligand_high}" if ref_flag else np.nan
        receptor_range = f"{receptor_low}-{receptor_high}" if ref_flag else np.nan


        if IS > up_IS:
            prediction.append(1)
            est_by_ref.append(True)
        elif IS < low_IS:
            prediction.append(0)
            est_by_ref.append(True)
        else:
            if ref_flag:
                if ref_sig:
                    if ligand_IS >= ligand_low and receptor_IS >= receptor_low:
                        prediction.append(1)
                        est_by_ref.append(True)
                    else:
                        prediction.append(1 if IS > null_IS else 0)
                        est_by_ref.append(False)
                else:
                    if ligand_IS <= ligand_high and receptor_IS <= receptor_high:
                        prediction.append(0)
                        est_by_ref.append(True)
                    else:
                        prediction.append(1 if IS > null_IS else 0)
                        est_by_ref.append(False)
            else:
                prediction.append(1 if IS > null_IS else 0)
                est_by_ref.append(False)

        if ref_flag:
            if prediction[-1] and ref_sig:
                comparison = "Both Sig"
            elif prediction[-1] and not ref_sig:
                comparison = "Up"
            elif not prediction[-1] and ref_sig:
                comparison = "Down"
            else:
                comparison = "Both NS"
        else:
            comparison = np.nan

        if debug_mode:
            results.append((
                index, is_in_ref, col,  ligand_gene_name, receptor_gene_name,
                f"{IS}", null_IS, f"{low_IS}-{up_IS}", L_null_ref, R_null_ref,
                True, ligand_perc, receptor_perc, 
                ref_perc, ref_ligand_perc, ref_receptor_perc, 
                ligand_IS, ligand_range, 
                receptor_IS, receptor_range,
                bool(prediction[-1]), ref_sig, comparison
            ))
        else:   
            results.append((
                index, is_in_ref, col,  ligand_gene_name, receptor_gene_name,
                f"{IS}", null_IS, f"{low_IS}-{up_IS}", 
                True, ligand_perc, receptor_perc, 
                ref_perc, ref_ligand_perc, ref_receptor_perc, 
                ligand_IS, ligand_range, 
                receptor_IS, receptor_range,
                bool(prediction[-1]), ref_sig, comparison
            ))
    
    for index, col in zip(*np.where(np.logical_and(ref_pvals < 0.05, percents_analysis.loc[common_ind, common_col] == False))):
        index = ref_pvals.index[index]
        col = ref_pvals.columns[col]

        #01-11 add
        if debug_mode == True:
            mid1 = interactions.loc[col].multidata_1_id
            mid2 = interactions.loc[col].multidata_2_id
            # L_null_ref = ref_mean_pmfs.iloc[0][mid1].mean
            # R_null_ref = ref_mean_pmfs.iloc[0][mid2].mean
            L_null_ref = null_ref_mean_counts[mid1].item()
            R_null_ref = null_ref_mean_counts[mid2].item()

        #01-11 end

        # 01-10 add
        sender, receiver = index.split('|')
        mid_1 = interactions.loc[col, 'multidata_1_id']
        mid_2 = interactions.loc[col, 'multidata_2_id']
        pmf_1 = mean_pmfs.loc[sender, mid_1]
        pmf_2 = mean_pmfs.loc[receiver, mid_2]
        pmf = (pmf_1 & pmf_2) / 2
        threshold = get_threshold_from_pmf(pmf, 0.05)
        up_IS = threshold + threshold / k * 1.96 
        low_IS = threshold - threshold / k * 1.96 
        # 01-10 end

        ligand_gene_name = interactions.loc[col].multidata_1_id
        receptor_gene_name = interactions.loc[col].multidata_2_id
        ligand_gene_name = id2symbol_dict[ligand_gene_name]
        receptor_gene_name =  id2symbol_dict[receptor_gene_name]

        IS = interactions_strength.loc[index, col]
        null_IS = null_interactions_strength.loc[index, col]
        ligand_perc = L_perc.loc[index, col]
        receptor_perc = R_perc.loc[index, col]
        ref_ligand_perc = ref_L_perc.loc[index, col]
        ref_receptor_perc = ref_R_perc.loc[index, col]
        ligand_IS = p1.loc[index, col]
        receptor_IS = p2.loc[index, col]

        ligand_low = np.max(ref_p1.loc[index, col] * (1-1/k*1.96), 0)
        ligand_high = ref_p1.loc[index, col] * (1+1/k*1.96)
        receptor_low = np.max(ref_p2.loc[index, col], 0) * (1-1/k*1.96)
        receptor_high = ref_p2.loc[index, col] * (1+1/k*1.96)
        ligand_range = f"{ligand_low}-{ligand_high}"
        receptor_range = f"{receptor_low}-{receptor_high}"

        if debug_mode:
            results.append((
                index, True, col, ligand_gene_name, 
                receptor_gene_name,
                f"{IS}", null_IS, f"{low_IS}-{up_IS}", L_null_ref, R_null_ref,
                False, ligand_perc, receptor_perc, 
                True, ref_ligand_perc, ref_receptor_perc, 
                ligand_IS, ligand_range, 
                receptor_IS, receptor_range,
                False, True, "Down"
            ))
        else:
            results.append((
                index, True, col, ligand_gene_name, 
                receptor_gene_name,
                f"{IS}", null_IS, f"{low_IS}-{up_IS}", 
                False, ligand_perc, receptor_perc, 
                True, ref_ligand_perc, ref_receptor_perc, 
                ligand_IS, ligand_range, 
                receptor_IS, receptor_range,
                False, True, "Down"
            ))

    if debug_mode:
        results_df = pd.DataFrame(
            results, 
            columns=[
                'sender|receiver', 'in_reference', 'LRI_ID', 
                'ligand', 'receptor', 'comm_score', 'null_comm_score',
                'sig_threshold_CI', 'ligand_null_ref', 'receptor_null_ref',
                'above_expr_threshold', 'ligand_expr_percent', 'receptor_expr_percent',
                'above_expr_threshold_ref', 'ligand_expr_percent_ref', 'receptor_expr_percent_ref',
                'ligand_CS_component', 'ligand_CS_CI',
                'receptor_CS_component', 'receptor_CS_CI',
                'is_significant', 'is_significant_ref', 'trend_vs_ref'
            ]
        )
    else:
        results_df = pd.DataFrame(
            results, 
            columns=[
                'sender|receiver', 'in_reference', 'LRI_ID', 
                'ligand', 'receptor', 'comm_score', 'null_comm_score',
                'sig_threshold_CI',
                'above_expr_threshold', 'ligand_expr_percent', 'receptor_expr_percent',
                'above_expr_threshold_ref', 'ligand_expr_percent_ref', 'receptor_expr_percent_ref',
                'ligand_CS_component', 'ligand_CS_CI',
                'receptor_CS_component', 'receptor_CS_CI',
                'is_significant', 'is_significant_ref', 'trend_vs_ref'
            ]
        )

    if debug_mode:
        logger.debug("Entering debug process")
        real_pvals = pd.read_csv(f'{save_path}/debug_pvals.txt', sep='\t', index_col=0)
        real_pvals = real_pvals.loc[interactions_strength.index, interactions_strength.columns]
        real_pvals_values = real_pvals.values[np.where(percents_analysis)]
        intersection = np.sum(np.logical_and(prediction, real_pvals_values < 0.05))
        union = np.sum(np.logical_or(prediction, real_pvals_values < 0.05))
        print(f"Intersection:{intersection}, Union:{union}")
        print(f"#Pred.Sig:{np.sum(prediction)}, #Real.Sig:{np.sum(real_pvals_values < 0.05)}, #valid:{len(real_pvals_values)}")
        print(f"Precision:{intersection/np.sum(prediction)}, IoU:{intersection/union}, Recall:{intersection/np.sum(real_pvals_values < 0.05)}")
        with open(f'{save_path}/debug_results.txt', 'wt') as f:
            f.write(f"Intersection:{intersection}, Union:{union}\n")
            f.write(f"#Pred.Sig:{np.sum(prediction)}, #Real.Sig:{np.sum(real_pvals_values < 0.05)}, #valid:{len(real_pvals_values)}\n")
            f.write(f"Precision:{intersection/np.sum(prediction)}, IoU:{intersection/union}, Recall:{intersection/np.sum(real_pvals_values < 0.05)}\n")

        est_by_ref = np.array(est_by_ref)
        print(f"#by_null: {np.sum(1 - est_by_ref)}, #by_ref: {np.sum(est_by_ref)}")
        real_pvals_values = real_pvals_values[est_by_ref]
        prediction = np.array(prediction)[est_by_ref]
        intersection = np.sum(np.logical_and(prediction, real_pvals_values < 0.05))
        union = np.sum(np.logical_or(prediction, real_pvals_values < 0.05))
        print(f"Real Prec:{intersection/np.sum(prediction)}, Real IoU:{intersection/union}, Real Recall:{intersection/np.sum(real_pvals_values < 0.05)}")
        with open(f'{save_path}/debug_results.txt', 'at') as f:
            f.write(f"#by_null: {np.sum(1 - est_by_ref)}, #by_ref: {np.sum(est_by_ref)}\n")
            f.write(f"Real Prec:{intersection/np.sum(prediction)}, Real IoU:{intersection/union}, Real Recall:{intersection/np.sum(real_pvals_values < 0.05)}")
        logger.success("Debug ends.")


    ####### save reference results #######
    logger.info("Saving inference results.")
    percents_analysis.to_csv(f'{save_path}/query_percents_analysis.tsv', sep='\t')
    interactions_strength.to_csv(f'{save_path}/query_interactions_strength.tsv', sep='\t')
    results_df.to_csv(f'{save_path}/query_infer_results.tsv', sep='\t', index=False)


def calculate_adjust_factor(query, reference_path, save_path, debug_mode=False):
    mean_hk_rnk, gene_index = record_hk_genes(query)
    query_hk = pd.DataFrame(np.array(mean_hk_rnk).flatten(), index=gene_index, columns=['query_hk'])
    ref_hk = pd.read_csv(f'{reference_path}/ref_hk.txt', sep='\t', index_col=0)
    hk_df = pd.merge(query_hk, ref_hk, left_index=True, right_index=True)

    mean_list = []
    std_list = []
    mean_by_std_list = []
    for i in np.arange(0,30):
        foo = np.where(np.logical_and(hk_df.ref_hk >=i, hk_df.ref_hk < i+5))
        mean_ = np.mean(np.array(hk_df.ref_hk)[foo])
        std_ = np.sqrt(np.mean(np.square(np.array(hk_df.query_hk)[foo] - np.array(hk_df.ref_hk)[foo])))
        mean_list.append(mean_)
        std_list.append(std_)
        mean_by_std_list.append(mean_/std_)
    k = np.nanmean(mean_by_std_list)
    logger.debug(f"k={k}")
    k = max(3, k)
    
    if debug_mode:
        import matplotlib.pyplot as plt 
        plt.figure(figsize=(6,3))
        plt.subplot(1,2,1)
        plt.scatter(hk_df.query_hk, hk_df.ref_hk, s=3, alpha=0.1)
        plt.subplot(1,2,2)
        plt.scatter(np.log(hk_df.query_hk), np.log(hk_df.ref_hk), s=3, alpha=0.1)
        plt.savefig(f'{save_path}/hk_scatter.jpg')

        plt.figure(figsize=(3,3))
        plt.scatter(mean_list, std_list)
        plt.title(np.nanmean(mean_by_std_list))
        plt.savefig(f'{save_path}/hk_mean_by_std.jpg')
        
    return k
    

def update_reference_for_first_time_activation(reference_path):
    with open(f'{reference_path}/basic_info_dict.pkl', 'rb') as f:
        basic_info_dict = pickle.load(f)
    
    from scipy.signal import fftconvolve
    from .distrib_digit import Distribution_digit
    
    n_bins = 50
    precision_digit = 0.01
    pmf_bins_digit = np.arange(0, n_bins+precision_digit - 1e-10, precision_digit)

    n_fft = 100
    gene_sum_pmf_dict = {}
    for gene in basic_info_dict:
        loc = basic_info_dict[gene]['loc'] 
        scale =  basic_info_dict[gene]['scale']
        gene_sum_pmf_dict[gene] = {1: basic_info_dict[gene]['expr_dist']}
        for item in range(2,n_fft):
            gene_sum_pmf_dict[gene][item] = fftconvolve(gene_sum_pmf_dict[gene][item-1], gene_sum_pmf_dict[gene][1])

    gene_pmf_dict = {}
    for gene in basic_info_dict:
        gene_pmf_dict[gene] = {}
        loc = basic_info_dict[gene]['loc']
        scale = basic_info_dict[gene]['scale']
        for item in range(1,n_fft):
            pmf = gene_sum_pmf_dict[gene][item]
            cdf = np.cumsum(pmf)
            pmf_array = np.diff(cdf[np.int64(pmf_bins_digit * item)],prepend=0)
            if item == 1:
                gene_pmf_dict[gene][item] = Distribution_digit('other', pmf_array=pmf_array, loc=loc, scale=scale, is_align=True)
            else:
                gene_pmf_dict[gene][item] = Distribution_digit('other', pmf_array=pmf_array, is_align=True)

    with open(f'{reference_path}/ref_gene_pmf_dict.pkl', 'wb') as f:
        pickle.dump(gene_pmf_dict, f)

    logger.success(f"Reference panel data updated.")


def infer_query_workflow(database_file_path, reference_path, query_counts_file_path, celltype_file_path, save_path, celltype_mapping_dict=None, meta_key=None, debug_mode=False):
    
    with open(f'{reference_path}/config.toml', 'rb') as f:
        config = tomllib.load(f)
        logger.info(f"Start inferring by using CCC reference: {config['reference_name']}")
        logger.info(f"Reference min_percentile = {config['min_percentile']}")
        logger.info(f"Reference LRI DB = {config['LRI_database']}")

    if not os.path.exists(f'{reference_path}/ref_gene_pmf_dict.pkl'):
        logger.info(f"Updating reference panel data for first time activation.")
        update_reference_for_first_time_activation(reference_path)

    if not os.path.exists(save_path):
        os.makedirs(save_path)
        logger.success(f"Query save dir is created.")
    else:
        logger.warning(f"{save_path} already exists, all files will be overwritten")


    query = sc.read_h5ad(query_counts_file_path)
    sc.pp.filter_cells(query, min_genes=50) # basic QC
    logger.info(f"Reading query adata, {query.shape[0]} cells x {query.shape[1]} genes")
    
    if meta_key is not None:
        labels_df = pd.DataFrame(query.obs[meta_key])
        labels_df.columns = ['cell_type']
        labels_df.index.name = 'barcode_sample'
    else:
        labels_df = pd.read_csv(celltype_file_path, sep='\t', index_col=0)
        for barcode in query.obs_names:
            assert barcode in labels_df.index, "The index of query data doesn't match the index of labels"
        labels_df = labels_df.loc[query.obs_names, :]

    query = rank_preprocess(query)
    if debug_mode:
        query.write_h5ad(f"{save_path}/debug_digit.h5ad")

    k = calculate_adjust_factor(query, reference_path, save_path, debug_mode)

    counts_df, complex_table, interactions = get_fastccc_input(query, database_file_path)
    
    if debug_mode:
        logger.debug("Entering debug process")
        from .build_reference import fastccc_for_reference
        fastccc_for_reference('', save_path, counts_df, labels_df, complex_table, interactions, min_percentile = config['min_percentile'], query_debug_mode=True)
        logger.debug("Debug ends.")

    compare_with_reference(
        counts_df, labels_df, complex_table, interactions, reference_path, save_path, 
        config=config, celltype_mapping_dict = celltype_mapping_dict, 
        database_file_path = database_file_path,
        k=k, debug_mode=debug_mode
    )
    logger.success("Inference workflow done.")


import scanpy as sc
import numpy as np
import pandas as pd
from scipy.sparse import issparse
from tqdm import tqdm
from loguru import logger
from .preprocess import get_interactions
from . import preproc_utils
from .core import calculate_cluster_percents, analyze_interactions_percents
from .distrib_digit import Distribution_digit, get_pmf_array_from_samples_for_digitized_bins,  get_minimum_distribution_for_digit
from . import dist_complex
from . import dist_lr
from . import score
import itertools
from scipy.signal import fftconvolve
from collections import Counter
import os
import pickle

def digitize_transform(x, n_bins=50):
    def _digitize(x: np.ndarray, bins: np.ndarray, side="both") -> np.ndarray:
        assert x.ndim == 1 and bins.ndim == 1
        left_digits = np.digitize(x, bins)
        if side == "one":
            return left_digits
        right_digits = np.digitize(x, bins, right=True)
        rands = np.random.rand(len(x))  # uniform random numbers
        digits = rands * (right_digits - left_digits) + left_digits
        digits = np.ceil(digits).astype(np.int64)
        return digits
        
    # non_zero_ids = x.nonzero()
    # non_zero_row = x[non_zero_ids]
    '''
    input x 就是 非0的，直接针对csr，coo数据的
    '''
    bins = np.quantile(x, np.linspace(0, 1, n_bins - 1))
    non_zero_digits = _digitize(x, bins)
    return non_zero_digits
    

def calculate_L_R_and_IS_score(mean_counts, interactions):
    p1_index = []
    p2_index = []
    all_index = []
    for i in itertools.product(sorted(mean_counts.index), sorted(mean_counts.index)):
        p1_index.append(i[0])
        p2_index.append(i[1])
        all_index.append('|'.join(i))
    p1 = mean_counts.loc[p1_index, interactions['multidata_1_id']]
    p2 = mean_counts.loc[p2_index, interactions['multidata_2_id']]
    p1.columns = interactions.index
    p2.columns = interactions.index
    p1.index = all_index
    p2.index = all_index
    interactions_strength = (p1 + p2)/2 * (p1 > 0) * (p2>0)
    return p1, p2, interactions_strength

def calculate_L_R_and_IS_percents(cluster_percents, interactions, threshold=0.1, sep='|'):
    
    p1_index = []
    p2_index = []
    all_index = []
    for i in itertools.product(sorted(cluster_percents.index), sorted(cluster_percents.index)):
        p1_index.append(i[0])
        p2_index.append(i[1])
        all_index.append(sep.join(i))
        
    p1 = cluster_percents.loc[p1_index, interactions['multidata_1_id']]
    p2 = cluster_percents.loc[p2_index, interactions['multidata_2_id']]
    p1.columns = interactions.index
    p2.columns = interactions.index
    p1.index = all_index
    p2.index = all_index
    
    interactions_strength = (p1>threshold) * (p2>threshold)
    # print((p1>threshold) * (p2>threshold))
    
    return p1, p2, interactions_strength

def calculate_mean_pmfs(counts_df, labels_df, complex_table, gene_pmf_dict, n_fft=100):
    meta_dict = Counter(labels_df.cell_type)
    ####### clusters_mean #######
    clusters_mean_dict = {}
    for celltype in sorted(meta_dict):
        clusters_mean_dict[celltype]  = {}
        n_sum = meta_dict[celltype]
        if n_sum < n_fft:
            for gene in counts_df.columns:
                if gene not in gene_pmf_dict:
                    continue
                else:
                    clusters_mean_dict[celltype][gene] = gene_pmf_dict[gene][n_sum]
        else:
            for gene in counts_df.columns:
                if gene not in gene_pmf_dict:
                    continue
                else:
                    clusters_mean_dict[celltype][gene] = gene_pmf_dict[gene][1] ** n_sum / n_sum
    mean_pmfs = pd.DataFrame(clusters_mean_dict).T
    complex_func = get_minimum_distribution_for_digit
    mean_pmfs = dist_complex.combine_complex_distribution_df(mean_pmfs, complex_table, complex_func)
    return mean_pmfs


def rank_preprocess(adata):
    np.random.seed(42) # add seed to ensure reproduiablity
    assert issparse(adata.X), "Anndata.X should be a sparse matrix format."
    if adata.shape[1] < 5000:
        logger.warning("Do you use whole transcriptomes? Raw data w\o filtering genes should work better.")

    for i in tqdm(range(adata.shape[0]), desc="Ranking genes for cells", unit="cell", 
              bar_format="{l_bar}{bar} | {n_fmt}/{total_fmt} cells completed", leave=False):
        indices = slice(adata.X.indptr[i], adata.X.indptr[i+1])
        x = adata.X.data[indices]
        adata.X.data[indices] = digitize_transform(x)
    logger.success("Rank preprocess done.")
    return adata

def get_fastccc_input(adata, lrdb_file_path, convert_type = 'hgnc_symbol'):
    logger.info("Loading LRIs database. hgnc_symbol as gene name is requested.")
    interactions = get_interactions(lrdb_file_path)
    ##### gene_table ########
    gene_table = pd.read_csv(os.path.join(lrdb_file_path, 'gene_table.csv'))
    protein_table = pd.read_csv(os.path.join(lrdb_file_path, 'protein_table.csv'))
    gene_table = gene_table.merge(protein_table, left_on='protein_id', right_on='id_protein')
    #########################


    ##### complex_table ######
    complex_composition = pd.read_csv(os.path.join(lrdb_file_path, 'complex_composition_table.csv'))
    complex_table = pd.read_csv(os.path.join(lrdb_file_path, 'complex_table.csv'))
    complex_table = complex_table.merge(complex_composition, left_on='complex_multidata_id', right_on='complex_multidata_id')

    # 让我们只关注 'complex_multidata_id'，'protein_multidata_id'
    complex_table = complex_table[['complex_multidata_id','protein_multidata_id']]
    '''
    complex_table(pandas.DataFrame):
    =======================================================
            | complex_multidata_id   |  protein_multidata_id
    -------------------------------------------------------
    0      |             1355       |          1134
    1      |             1356       |          1175
    2      |             1357       |          1167
    =======================================================
    '''
    ##########################

    ##### feature to id conversion  ######
    # 不在 标准列表 里的 gene 就不要了
    tmp = gene_table[[convert_type, 'protein_multidata_id']]
    tmp = tmp.drop_duplicates()
    tmp.set_index('protein_multidata_id', inplace=True)

    select_columns = []
    columns_names = []
    for foo, boo in zip(tmp.index, tmp[convert_type]):
        if boo in adata.var_names:#counts.columns:
            select_columns.append(boo)
            columns_names.append(foo)

    reduced_counts = adata[:, select_columns].to_df()
    reduced_counts.columns = columns_names
    reduced_counts = reduced_counts.T.groupby(reduced_counts.columns).mean().T
    # FutureWarning: DataFrame.groupby with axis=1 is deprecated. Do `frame.T.groupby(...)` without axis instead.
    # reduced_counts = reduced_counts.groupby(reduced_counts.columns, axis=1).mean()
    ######################################

    ########## filter genes  ############
    # gene 在 所有 cell 上为 0 不要
    reduced_counts = preproc_utils.filter_empty_genes(reduced_counts)
    ######################################
    
    ######################################################################
    #                      3.Other DF filtered                       #
    ######################################################################
    # 一个 interaction 可能只有 partA 存在，但是 partB 不存在
    # 只有 如果 任意一部分不存在， 另一部分没必要参与后续计算

    ##### delete item not involved interactions ####

    foo_dict = complex_table.groupby('complex_multidata_id').apply(lambda x: list(x['protein_multidata_id'].values), include_groups=False).to_dict()
    '''
    dictionary complex_id: [protein_id_1, pid2, pid3, ...]
    foo_dict = {
        1355: [1134],
        1356: [1175],
        xxxx: [AAAA, BBBB, CCCC],
    }
    '''

    def __content__(key):
        if key not in foo_dict:
            return [key]
        else:
            return foo_dict[key]

    def __exist__(key, df):
        # 目前的 complex 策略就是 全部都要有
        # 经测试，这是cpdb用的策略
        for item in __content__(key):
            if item not in df.columns:
                return False
        return True

    temp_list = []
    temp_dict = {}
    for item in reduced_counts.columns:
        temp_dict[item] = False

    # 注释是为了验证 interactions 的过滤策略， 完全一致
    # print(interactions)
    select_index = []
    for partA, partB in zip(interactions.multidata_1_id, interactions.multidata_2_id):
        if __exist__(partA, reduced_counts) and __exist__(partB, reduced_counts):
            temp_list.extend([partA, partB])
            select_index.append(True)
        else:
            select_index.append(False)
    interactions_filtered = interactions[select_index]

    for item in temp_list:
        for subitem in __content__(item):
            if subitem in temp_dict:
                temp_dict[subitem] = True
    select_index = [key for key in temp_dict if temp_dict[key]]
    reduced_counts = reduced_counts[select_index]

    counts_df = reduced_counts
    temp_list = set(temp_list)
    select_index = [True if item in temp_list else False for item in complex_table.complex_multidata_id]
    complex_table = complex_table[select_index]
    interactions = interactions_filtered
    logger.success("Requested data for fastccc is prepared.")
    return counts_df, complex_table, interactions


def fastccc_for_reference(reference_name, save_path, counts_df, labels_df, complex_table, interactions, min_percentile = 0.1, ref_debug_mode=False, query_debug_mode=False, for_uploading=False):
    logger.info("Running FastCCC.")
    mean_counts = score.calculate_cluster_mean(counts_df, labels_df)
    complex_func = score.calculate_complex_min_func
    mean_counts = score.combine_complex_distribution_df(mean_counts, complex_table, complex_func)
    percents = calculate_cluster_percents(counts_df, labels_df, complex_table)

    n_bins = 50
    precision_digit = 0.01
    pmf_bins_digit = np.arange(0, n_bins+precision_digit - 1e-10, precision_digit)

    ####### 
    logger.info("Calculating null distributions.")
    n_fft = 100
    gene_sum_pmf_dict = {}
    basic_info_dict = {}
    for gene in counts_df.columns:
        samples = counts_df[gene].values

        loc = np.mean(samples)
        scale = np.std(samples)
        basic_info_dict[gene] = {'loc':loc, 'scale':scale}

        gene_sum_pmf_dict[gene] = {1: get_pmf_array_from_samples_for_digitized_bins(samples)}
        basic_info_dict[gene]['expr_dist'] = gene_sum_pmf_dict[gene][1]

        for item in range(2,n_fft):
            gene_sum_pmf_dict[gene][item] = fftconvolve(gene_sum_pmf_dict[gene][item-1], gene_sum_pmf_dict[gene][1])

    gene_pmf_dict = {}
    for gene in counts_df.columns:
        gene_pmf_dict[gene] = {}
        loc = basic_info_dict[gene]['loc']
        scale = basic_info_dict[gene]['scale']
        for item in range(1,n_fft):
            pmf = gene_sum_pmf_dict[gene][item]
            cdf = np.cumsum(pmf)
            pmf_array = np.diff(cdf[np.int64(pmf_bins_digit * item)],prepend=0)
            if item == 1:
                gene_pmf_dict[gene][item] = Distribution_digit('other', pmf_array=pmf_array, loc=loc, scale=scale, is_align=True)
            else:
                gene_pmf_dict[gene][item] = Distribution_digit('other', pmf_array=pmf_array, is_align=True)

    if ref_debug_mode or query_debug_mode:
        p1_index = []
        p2_index = []
        all_index = []
        for i in itertools.product(sorted(mean_counts.index), sorted(mean_counts.index)):
            p1_index.append(i[0])
            p2_index.append(i[1])
            all_index.append('|'.join(i))
        p1 = mean_counts.loc[p1_index, interactions['multidata_1_id']]
        p2 = mean_counts.loc[p2_index, interactions['multidata_2_id']]
        p1.columns = interactions.index
        p2.columns = interactions.index
        p1.index = all_index
        p2.index = all_index
        interactions_strength = (p1 + p2)/2 * (p1 > 0) * (p2>0)
        L_perc, R_perc, percents_analysis = calculate_L_R_and_IS_percents(percents, interactions, threshold=min_percentile)
        
        meta_dict = Counter(labels_df.cell_type)
        ####### clusters_mean #######
        clusters_mean_dict = {}
        for celltype in sorted(meta_dict):
            clusters_mean_dict[celltype]  = {}
            n_sum = meta_dict[celltype]
            if n_sum < n_fft:
                for gene in counts_df.columns:
                    clusters_mean_dict[celltype][gene] = gene_pmf_dict[gene][n_sum]
            else:
                for gene in counts_df.columns:
                    clusters_mean_dict[celltype][gene] = gene_pmf_dict[gene][1] ** n_sum / n_sum
        mean_pmfs = pd.DataFrame(clusters_mean_dict).T
        complex_func = get_minimum_distribution_for_digit
        mean_pmfs = dist_complex.combine_complex_distribution_df(mean_pmfs, complex_table, complex_func)

        logger.info("Calculating sig. LRIs.")
        pvals = dist_lr.calculate_key_interactions_pvalue(
            mean_pmfs, interactions, interactions_strength, percents_analysis, method='Arithmetic'
        )

        if query_debug_mode:
            pvals.to_csv(f'{save_path}/debug_pvals.txt', sep='\t')
            return

        if ref_debug_mode:
            pvals.to_csv(f'{save_path}/ref_pvals.txt', sep='\t')
            percents_analysis.to_csv(f'{save_path}/ref_percents_analysis.txt', sep='\t')
            L_perc.to_csv(f'{save_path}/ref_percents_L.txt', sep='\t')
            R_perc.to_csv(f'{save_path}/ref_percents_R.txt', sep='\t')
            # interactions_strength.to_csv(f'{save_path}/ref_interactions_strength.csv')
            p1.to_csv(f'{save_path}/ref_interactions_strength_L.txt', sep='\t')
            p2.to_csv(f'{save_path}/ref_interactions_strength_R.txt', sep='\t')

    ####### save reference results #######
    logger.info("Saving reference.")
    if for_uploading:
        with open(f'{save_path}/basic_info_dict.pkl', 'wb') as f:
            pickle.dump(basic_info_dict, f)
    else:
        with open(f'{save_path}/ref_gene_pmf_dict.pkl', 'wb') as f:
            pickle.dump(gene_pmf_dict, f)
    with open(f'{save_path}/ref_percents.pkl', 'wb') as f:
        pickle.dump(percents, f)
    with open(f'{save_path}/ref_mean_counts.pkl', 'wb') as f:
        pickle.dump(mean_counts, f)
    with open(f'{save_path}/complex_table.pkl', 'wb') as f:
        pickle.dump(complex_table, f)
    with open(f'{save_path}/interactions.pkl', 'wb') as f:
        pickle.dump(interactions, f)


def record_hk_genes(adata):
    from .hk_genes import housekeeping_genes
    select_index = [item for item in adata.var_names if item in housekeeping_genes]
    hk_adata = adata[:, select_index]
    mean_hk_rnk = hk_adata.X.mean(axis=0)
    return mean_hk_rnk, select_index

def record_adjustment_info(adata, save_path):
    mean_hk_rnk, gene_index = record_hk_genes(adata)
    ref_hk = pd.DataFrame(np.array(mean_hk_rnk).flatten(), index=gene_index, columns=['ref_hk'])
    ref_hk.to_csv(f'{save_path}/ref_hk.txt', sep='\t')


reference_config = {}

from datetime import date

def dumps(toml_dict, table=""):
    document = []
    for key, value in toml_dict.items():
        match value:
            case dict():
                table_key = f"{table}.{key}" if table else key
                document.append(
                    f"\n[{table_key}]\n{_dumps_dict(value)}"
                )
            case _:
                document.append(f"{key} = {_dumps_value(value)}")
    return "\n".join(document)

def _dumps_dict(toml_dict):
    document = []
    for key, value in toml_dict.items():
        key = f'"{key}"'
        document.append(f"{key} = {_dumps_value(value)}")
    return "\n".join(document)

def _dumps_value(value):
    match value:
        case bool():
            return "true" if value else "false"
        case float() | int():
            return str(value)
        case str():
            return f'"{value}"'
        case date():
            return value.isoformat()
        case list():
            return f"[{', '.join(_dumps_value(v) for v in value)}]"
        case _:
            raise TypeError(
                f"{type(value).__name__} {value!r} is not supported"
            )

def save_config(save_path):
    logger.info("Saving reference config.")
    save_content = dumps(reference_config)
    with open(f'{save_path}/config.toml', 'w') as f:
        f.write(save_content) 


def build_reference_workflow(database_file_path, reference_counts_file_path, celltype_file_path, reference_name, save_path, meta_key=None, min_percentile = 0.1, debug_mode=False, for_uploading=False):
    logger.info(f"Start building CCC reference.")

    reference_config['reference_name'] = reference_name
    reference_config['min_percentile'] = min_percentile
    if database_file_path.endswith('/'):
        reference_config['LRI_database'] = database_file_path[:-1].split('/')[-1]
    else:
        reference_config['LRI_database'] = database_file_path.split('/')[-1]

    logger.info(f"Reference_name = {reference_config['reference_name']}")
    logger.info(f"min_percentile = {reference_config['min_percentile']}")
    logger.info(f"LRI database = {reference_config['LRI_database']}")

    save_path = os.path.join(save_path, reference_name)
    if not os.path.exists(save_path):
        os.makedirs(save_path)
        logger.success(f"Reference save dir {save_path} is created.")
    else:
        logger.warning(f"{save_path} already exists, all files will be overwritten")

    reference = sc.read_h5ad(reference_counts_file_path)
    sc.pp.filter_cells(reference, min_genes=50)
    logger.info(f"Reading reference adata, {reference.shape[0]} cells x {reference.shape[1]} genes.")

    if meta_key is not None:
        labels_df = pd.DataFrame(reference.obs[meta_key])
        labels_df.columns = ['cell_type']
        labels_df.index.name = 'barcode_sample'
    else:
        labels_df = pd.read_csv(celltype_file_path, sep='\t', index_col=0)
        for barcode in reference.obs_names:
            assert barcode in labels_df.index, "The index of query data doesn't match the index of labels"
        labels_df = labels_df.loc[reference.obs_names, :]
    
    ct_counter = Counter(labels_df['cell_type'])
    reference_config['celltype'] = ct_counter
    
    reference = rank_preprocess(reference)
    record_adjustment_info(reference, save_path)
    counts_df, complex_table, interactions = get_fastccc_input(reference, database_file_path)
    fastccc_for_reference(reference_name, save_path, counts_df, labels_df, complex_table, interactions, min_percentile, ref_debug_mode=debug_mode, for_uploading=for_uploading)
    save_config(save_path)
    logger.success(f"Reference '{reference_name}' is built.")




import copy
import numpy as np
import pandas as pd
from .distrib import Distribution, get_norm_minimum_cdf_func, pmf_bins

def combine_complex_distribution_df(set_distribution_df, complex_table, complex_process_function):
    def sub_func(x):
        return complex_process_function(*x)

    def func(x):
        x = [sub_x for sub_x in x if sub_x in set_distribution_df.columns]
        if len(x) == 0:
            # 如果出现这种情况，其实应该报错，前面数据处理应该排除这种情况。
            return pd.Series(index=mean_pmf.index)
        return set_distribution_df.loc[:,x].apply(sub_func, axis=1)

    if not complex_table.empty:
        complex_pmf = complex_table.groupby('complex_multidata_id').apply(lambda x: x['protein_multidata_id'].values, include_groups=False).apply(lambda x:func(x)).T
        complex_pmf = complex_pmf.dropna(axis=1)
        set_distribution_df = pd.concat((set_distribution_df, complex_pmf), axis=1)
        
    return set_distribution_df


def get_minimum_distribution(*pmf_list):
    '''
    distribution_list = k distribution 
    F = k * len(distribuion) matrix (pmf -> cdf use np.cumsum)
    minimum F(x) = 1 - \\product (1 - F_i(x))
    so f(x) = np.diff(F(x))
    '''
    for pmf in pmf_list:
        assert pmf.is_align, "PMF is not aligned"
    if len(pmf_list) == 1:
        return pmf_list[0]
    
    ### update 0203 ###
    loc_list, scale_list = [], []
    flag = True
    for pmf in pmf_list:
        if pmf.is_normal_type:
            loc_list.append(pmf.loc)
            scale_list.append(pmf.scale)
        else:
            flag = False
            break
    if flag:
        
        ### update 0204 ###
        arg_loc = np.argsort(loc_list)
        loc_min = loc_list[arg_loc[0]]
        loc_other = np.array(loc_list)[arg_loc[1:]]
        scale_min = scale_list[arg_loc[0]]
        scale_other = np.array(scale_list)[arg_loc[1:]]
        if loc_min + 5* scale_min <= np.min(loc_other - 5*scale_other):
            return Distribution(dtype='normal', loc=loc_min, scale=scale_min)
        
        ### ### ### ### ###
        
        analytic_func = get_norm_minimum_cdf_func(locs = loc_list, scales=scale_list)
        f = np.diff(analytic_func(pmf_bins), prepend=0)  # 没矫正和为1
        pmf = Distribution(dtype='other', pmf_array=f, is_align=True)
        pmf.is_analytic = True
        pmf.cdf_analytic_func = analytic_func
        return pmf
    ### ### ### ### ###
    
    distribution_list = [pmf.get_pmf_array() for pmf in pmf_list]
    max_length = max(arr.shape[0] for arr in distribution_list)
    padded_arrays = [np.pad(arr, (0, max_length - arr.shape[0]), 'constant') for arr in distribution_list]
    F = np.vstack(padded_arrays)
    # F = np.stack(distribution_list)
    F = np.cumsum(F, axis=1)
    minF = 1 - np.prod(1-F, axis=0)
    f = np.diff(minF, prepend=0)
    return Distribution(dtype='other', pmf_array=f, is_align=True)


def get_average_distribution(*distribution_list):

    assert len(distribution_list) > 0, "Distribution list is empty."
    for distribution in distribution_list:
        assert distribution.is_align, "PMF is not aligned"
    if len(distribution_list) == 1:
        return distribution_list[0]
    
    new_distribution = distribution_list[0]
    for i in range(1, len(distribution_list)):
        new_distribution += distribution_list[i]
    new_distribution /= len(distribution_list)
    return new_distribution


import os
import numpy as np
import pandas as pd
from scipy.special import loggamma, gamma
from scipy.optimize import minimize

from scipy.signal import fftconvolve
from scipy.stats import norm

from typing import Literal, Union
from functools import lru_cache

from tqdm import tqdm
from loguru import logger

n_bins = 50
precision = 0.01
n_fft = 100
pmf_bins = np.arange(0, n_bins+precision - 1e-10, precision)
pmf_int_bins = np.arange(len(pmf_bins))


def get_norm_cdf_func(loc, scale):
    def analytic_func(x):
        return norm.cdf(x, loc, scale)
    return analytic_func

def get_norm_minimum_cdf_func(locs, scales):
    def analytic_func(x, locs=locs, scales=scales):
        F = []
        for loc, scale in zip(locs, scales):
            F.append(norm.cdf(x, loc, scale))
        minF = 1 - np.prod(1-np.array(F), axis=0)
        return minF
    return analytic_func


class Distribution_digit:
    def __init__(
        self, 
        dtype: Literal['normal', 'gaussian', 'other'] = 'other', 
        loc: Union[float, None] = None, 
        scale: Union[float, None] = None, 
        samples: Union[np.ndarray, None] = None,
        pmf_array: Union[np.ndarray, None] = None, 
        is_align: Union[bool, None] = None,
        eps: float = 1e-4
    ):

        if dtype in ['normal', 'gaussian']:
            assert loc is not None and scale is not None
            is_align = True
            dtype = 'normal'
        else:
            if samples is not None:
                pmf_array = get_pmf_array_from_samples(samples)
                is_align = True
            assert pmf_array is not None and is_align is not None
            pmf_array = np.trim_zeros(pmf_array, trim='b')
            assert np.abs(1 - np.sum(pmf_array)) < eps, f"The sum of PMF is {np.sum(pmf_array)} not 1."
            # if np.abs(1 - np.sum(pmf_array)) >= 1e-7:
            #     print(f"The sum of PMF is {np.sum(pmf_array)} not 1.")
            
        self.loc = loc
        self.scale = scale
        self.dtype = dtype
        self.is_align = is_align
        self.pmf_array = pmf_array
        self.is_normal_type = self.dtype == 'normal'
        self.is_analytic = self.is_normal_type # normal or combine of normal
        self.is_complex_analytic = False
        if self.is_normal_type:
            self.cdf_analytic_func = get_norm_cdf_func(loc, scale)
            self.min_cdf_non_zero = self.loc - 5 * self.scale
            self.min_cdf_one = self.loc + 5 * self.scale
            self.support_length = self.min_cdf_one - self.min_cdf_non_zero
            
        # supplementary property
        self.cdf_array = None
    
    @property
    def mean(self):
        return self.get_mean()
    
    @property
    def std(self):
        return self.get_std()
        
    @property    
    def var(self):
        if self.is_normal_type:
            return np.square(self.scale)
        return np.sum(np.square(np.arange(len(self.pmf_array)) * precision - self.mean) 
                      * self.pmf_array )
    @property
    def pmf(self):
        return self.get_pmf_array()
    
    @property
    def cdf(self):
        return self.get_cdf_array()
        
    def get_mean(self):
        if self.loc is None:
            self.loc = np.sum(self.pmf_array * np.arange(len(self.pmf_array))) * precision
        return self.loc
    
    def get_std(self):
        if self.scale is None:
            self.scale = np.sqrt(self.var)
        return self.scale
            
    def get_pmf_array(self):
        if self.pmf_array is None: # which is a normal distribution
            # >>>>>>>>>>> update precision >>>>>>>>>> #
            # old version:
            # pdf = norm.pdf(pmf_bins, loc=self.loc, scale=self.scale)
            # self.pmf_array = pdf / np.sum(pdf)
            # new version: 
            cdf = norm.cdf(pmf_bins, loc=self.loc, scale=self.scale)
            self.pmf_array = np.diff(cdf, prepend=0)
            # <<<<<<<<<<< update precision <<<<<<<<<< #
            self.pmf_array = np.trim_zeros(self.pmf_array, trim='b')
        return self.pmf_array
        
    def get_cdf_array(self):
        if self.cdf_array is None:
            self.cdf_array = np.cumsum(self.get_pmf_array())
        return self.cdf_array
        
    def __add__(self, pmf2):
        if self.is_normal_type and pmf2.is_normal_type:
            new_loc = self.loc + pmf2.loc
            new_scale = np.sqrt(np.square(self.scale) + np.square(pmf2.scale))
            return Distribution_digit(dtype='normal', loc=new_loc, scale=new_scale)
        
        new_pmf_array = fftconvolve(self.get_pmf_array(), pmf2.get_pmf_array())
        new_pmf = Distribution_digit(dtype='other', pmf_array=new_pmf_array, is_align=False)
        
        # 为了精确 打得补丁
        if self.is_analytic and pmf2.is_analytic:
            self.is_complex_analytic = True
            new_pmf.ligand = self
            new_pmf.receptor = pmf2
        ###
 
        return new_pmf
        
    def __pow__(self, n):
        assert type(n) == int
        assert n >= 1
        if n == 1:
            return self
        
        if n >= n_fft or self.is_normal_type:
            new_loc = self.loc * n
            new_scale = np.sqrt(n) * self.scale
            return Distribution_digit(dtype='normal', loc=new_loc, scale=new_scale)
        
        else:
            new_pmf_array = self.get_pmf_array()
            for i in range(1, n):
                new_pmf_array = fftconvolve(new_pmf_array, self.get_pmf_array())
                # new_pmf_array = np.trim_zeros(new_pmf_array, trim='b') 应该是不会进入这个优化，因为前面都是有非0，卷积仍然非0，除非太小？
            return Distribution_digit(dtype='other', pmf_array=new_pmf_array, is_align=False)

    def __truediv__(self, n):
        assert type(n) == int
        assert n >= 1
        if n == 1:
            return self
        
        if self.is_normal_type:
            new_loc = self.loc / n
            new_scale = self.scale / n
            return Distribution_digit(dtype='normal', loc=new_loc, scale=new_scale)

        cdf = self.get_cdf_array()
        bins = pmf_int_bins * n
        n_cut = np.sum(bins < len(cdf))
        # 增加 append 1.0，修正可能的bug
        average_pmf_array = np.diff(np.append(cdf[bins[:n_cut]], 1.0), prepend=0)
        average_pmf =  Distribution_digit(dtype='other', pmf_array=average_pmf_array, is_align=True)
        
        # 为了精确 打得补丁
        if self.is_complex_analytic:
            average_pmf.is_complex_analytic = True
            average_pmf.ligand = self.ligand
            average_pmf.receptor = self.receptor
        ####
            
        return average_pmf

    def __and__(self, pmf2):
        if self.is_normal_type and pmf2.is_normal_type:
            return self.__add__(pmf2)
        p1 = self.get_pmf_array()
        p2 = pmf2.get_pmf_array()
        pr0 = p1[0] + p2[0] - p1[0]*p2[0]
        new_pmf_array = fftconvolve(p1[1:], p2[1:])
        new_pmf_array = np.insert(new_pmf_array, 0, [pr0, 0])
        return Distribution_digit(dtype='other', pmf_array=new_pmf_array, is_align=False)



def get_pmf_array_from_samples_for_digitized_bins(samples, n_bins=50):
    assert type(n_bins) == int and n_bins > 0, "n_bins should be positive integer"
    assert np.max(samples) <= n_bins, "error: n_bins < max(samples)"
    pmf_array, _ = np.histogram(samples, np.arange(n_bins+2))
    pmf_array = pmf_array / np.sum(pmf_array)
    return pmf_array


def get_minimum_distribution_for_digit(*pmf_list):
    '''
    distribution_list = k distribution 
    F = k * len(distribuion) matrix (pmf -> cdf use np.cumsum)
    minimum F(x) = 1 - \\product (1 - F_i(x))
    so f(x) = np.diff(F(x))
    '''

    for pmf in pmf_list:
        assert pmf.is_align, "PMF is not aligned"
    if len(pmf_list) == 1:
        return pmf_list[0]
    
    ### update 0203 ###
    loc_list, scale_list = [], []
    flag = True
    for pmf in pmf_list:
        if pmf.is_normal_type:
            loc_list.append(pmf.loc)
            scale_list.append(pmf.scale)
        else:
            flag = False
            break
    if flag:
        
        ### update 0204 ###
        arg_loc = np.argsort(loc_list)
        loc_min = loc_list[arg_loc[0]]
        loc_other = np.array(loc_list)[arg_loc[1:]]
        scale_min = scale_list[arg_loc[0]]
        scale_other = np.array(scale_list)[arg_loc[1:]]
        if loc_min + 5* scale_min <= np.min(loc_other - 5*scale_other):
            return Distribution_digit(dtype='normal', loc=loc_min, scale=scale_min)
        
        ### ### ### ### ###
        
        analytic_func = get_norm_minimum_cdf_func(locs = loc_list, scales=scale_list)
        f = np.diff(analytic_func(pmf_bins), prepend=0)  # 没矫正和为1
        pmf = Distribution_digit(dtype='other', pmf_array=f, is_align=True)
        pmf.is_analytic = True
        pmf.cdf_analytic_func = analytic_func
        return pmf
    ### ### ### ### ###
    
    distribution_list = [pmf.get_pmf_array() for pmf in pmf_list]
    max_length = max(arr.shape[0] for arr in distribution_list)
    padded_arrays = [np.pad(arr, (0, max_length - arr.shape[0]), 'constant') for arr in distribution_list]
    F = np.vstack(padded_arrays)
    # F = np.stack(distribution_list)
    F = np.cumsum(F, axis=1)
    minF = 1 - np.prod(1-F, axis=0)
    f = np.diff(minF, prepend=0)
    return Distribution_digit(dtype='other', pmf_array=f, is_align=True)


import pandas as pd
import numpy as np
import itertools
import timeit
from .distrib import get_pvalue_from_pmf

def calculate_key_interactions_pvalue(mean_pmf, interactions, interactions_strength, percent_analysis, method='Arithmetic'):
    p1_index = []
    p2_index = []
    all_index = []
    for i in itertools.product(sorted(mean_pmf.index), sorted(mean_pmf.index)):
        p1_index.append(i[0])
        p2_index.append(i[1])
        all_index.append('|'.join(i))
        
    p1 = mean_pmf.loc[p1_index, interactions['multidata_1_id']]
    p2 = mean_pmf.loc[p2_index, interactions['multidata_2_id']]
    p1.columns = interactions.index
    p2.columns = interactions.index
    p1.index = all_index
    p2.index = all_index
    
    p1_items = p1.values[np.where(percent_analysis)]
    p2_items = p2.values[np.where(percent_analysis)]
    
    
    start = timeit.default_timer()
    if method == 'Arithmetic':
        pval_pmfs = (p1_items & p2_items) / 2
    elif method == 'Geometric':
        pval_pmfs = p1_items * p2_items
    stop = timeit.default_timer()
    
    mean_gt = interactions_strength.values[np.where(percent_analysis)]
    est = []
    for i, value in enumerate(mean_gt):
        pval_est = get_pvalue_from_pmf(value, pval_pmfs[i])
        est.append(pval_est)
        
    pvalues = np.ones_like(interactions_strength)
    pvalues[np.where(percent_analysis)] = est
    pvalues = pd.DataFrame(pvalues, index=interactions_strength.index, columns=interactions_strength.columns)
    return pvalues


import sys
from loguru import logger

logger.remove()
logger.add(sys.stdout, level="INFO", format='<cyan>{time:YYYY-MM-DD HH:mm:ss}</cyan> | <level>{level: <8}</level> | <level>{message}</level>')

from .core import Cauchy_combination_of_statistical_analysis_methods
from .core import statistical_analysis_method


import sys
import os, glob
import math
import getopt
import numpy as np
import pandas as pd
import pickle as pkl
from loguru import logger

def cauthy_combine(fastCCC_dir, task_id=None):
    if task_id is None:
        logger.warning("No task_id is provided, all pvals files will be combined.")
        pval_paths = glob.glob(fastCCC_dir+os.sep+'*pvals.tsv')
    else:
        logger.info(f"Task ID for combining is :{task_id}")
        pval_paths = glob.glob(fastCCC_dir+os.sep+f'{task_id}*pvals.tsv')
    logger.info(f"There are {len(pval_paths)} pval files.")
    joined_path = '\n'.join(pval_paths)
    logger.debug(f"\n{joined_path}")

    ct_pairs, cpis = None, None
    comb_dict = dict()
    for pval_path in pval_paths:
        comb = os.path.basename(pval_path)
        comb = comb.replace('pvals.tsv', '')
        pval_df = pd.read_csv(pval_path, header=0, index_col=0, sep='\t')
        if ct_pairs is None:
            ct_pairs = pval_df.index.tolist()
        if cpis is None:
            cpis = pval_df.columns.tolist()
        comb_dict[comb] = pval_df.values

    pval_mat = []
    for comb, values in comb_dict.items():
        pval_mat.append(np.expand_dims(values, axis=1))
    pval_mat = np.concatenate(pval_mat, axis=1)
    weight = np.ones(len(comb_dict)) / len(comb_dict)
    T = pval_mat.copy()
    T[np.where(pval_mat == 1)] = np.tan(-np.pi/2)
    T[foo] = np.tan(np.pi*(0.5 - T[(foo:=np.where(pval_mat != 1))]))
    T =  weight @ T
    P = 0.5 - np.arctan(T) / np.pi

    T_df = pd.DataFrame(T, index=ct_pairs, columns=cpis)
    P_df = pd.DataFrame(P, index=ct_pairs, columns=cpis)

    if task_id is None:
        T_df.to_csv(fastCCC_dir+os.sep+'Cauchy_stats.tsv', sep='\t')
        P_df.to_csv(fastCCC_dir+os.sep+'Cauchy_pvals.tsv', sep='\t')
    else:
        T_df.to_csv(fastCCC_dir+os.sep+f'{task_id}_Cauchy_stats.tsv', sep='\t')
        P_df.to_csv(fastCCC_dir+os.sep+f'{task_id}_Cauchy_pvals.tsv', sep='\t')

# if __name__ == "__main__":
#     cauthy_combine(fastCCC_dir)


import numpy as np
import pandas as pd
import os
import psutil


def create_significant_interactions_df(
    pvals, 
    LRI_db_path, 
    save = False, 
    save_path = './temp/', 
    save_file = 'significant_interaction_list', 
    seperator = '|'
):

    from . import preprocess
    interactions = preprocess.get_interactions(LRI_db_path)
    id2symbol_dict = (pd.read_csv(f'{LRI_db_path}/gene_table.csv')[['protein_id', 'hgnc_symbol']]\
    .set_index('protein_id')['hgnc_symbol']).to_dict()
    ##### complex_table ######
    complex_composition = pd.read_csv(os.path.join(LRI_db_path, 'complex_composition_table.csv'))
    complex_table = pd.read_csv(os.path.join(LRI_db_path, 'complex_table.csv'))
    complex_table = complex_table.merge(complex_composition, left_on='complex_multidata_id', right_on='complex_multidata_id')
    foo_dict = complex_table.groupby('complex_multidata_id').apply(lambda x: list(x['protein_multidata_id'].values), include_groups=False).to_dict()
    for key in foo_dict:
        value = ','.join([id2symbol_dict[item] for item in foo_dict[key]])
        id2symbol_dict[key] = value

    x, y = np.where(pvals.values<0.05)
    lines = []
    for subx, suby in zip(x,y):
        celltype_A, celltype_B = pvals.index[subx].split(seperator)
        interaction_ID = pvals.columns[suby]
        lines.append((celltype_A, celltype_B, interaction_ID, pvals.iloc[subx, suby]))

    output_df = pd.DataFrame(lines, columns=['sender_celltype', 'receiver_celltype', 'LRI_ID', 'p-value'])
    output_df = output_df.merge(interactions, left_on='LRI_ID', right_index=True, how='left')

    output_df.multidata_1_id = [id2symbol_dict[ligand_gene_name] for ligand_gene_name in output_df.multidata_1_id.tolist()]
    output_df.multidata_2_id = [id2symbol_dict[receptor_gene_name] for receptor_gene_name in output_df.multidata_2_id.tolist()]

    output_df = output_df[['sender_celltype', 'receiver_celltype', 'LRI_ID', 'multidata_1_id', 'multidata_2_id', 'p-value']]
    output_df = output_df.rename(columns={'multidata_1_id': 'ligand', 'multidata_2_id': 'receptor'})

    save_file = os.path.join(save_path, save_file)
    if save:
        output_df.to_excel(f'{save_file}.xlsx')
    return output_df


def create_significant_interactions_with_flag_df(pvals, significant_flag, interactions, save_path='./temp/', save_file='significant_interaction_list'):
    seperator = '|'
    x, y = np.where(pvals.values<0.05)
    lines = []
    for subx, suby in zip(x,y):
        if not significant_flag.iloc[subx, suby]:
            continue
        celltype_A, celltype_B = pvals.index[subx].split(seperator)
        interaction_ID = pvals.columns[suby]
        lines.append((celltype_A, celltype_B, interaction_ID, pvals.iloc[subx, suby]))
    output_df = pd.DataFrame(lines, columns=['Ligand_celltype', 'Receptor_celltype', 'Interaction_ID', 'P-val'])
    output_df = output_df.merge(interactions, left_on='Interaction_ID', right_index=True, how='left')
    save_file = os.path.join(save_path, save_file)
    output_df.to_excel(f'{save_file}.xlsx')
    return output_df

def get_current_memory():
    """
    获取当前内存占用
    usage:
    current_memory = get_current_memory()
    print("当前内存占用: {:.2f} MB".format(current_memory))
    """
    process = psutil.Process()
    return process.memory_info().rss / (1024 * 1024)  # 转换为MB



import pandas as pd
import numpy as np
import itertools
from functools import partial


def calculate_cluster_mean(counts_df, labels_df):
    '''
    input:
        counts_df(pandas.DataFrame): 
            sample * feature matrix. 
            The expression level of every gene in every cell.
            i.e.:
                =======================================
                      | gene1 | gene2 | gene3 | gene4 
                =======================================
                cell1 |  0.4  |  0.1  |  0.8  |  0.0
                cell2 |  0.0  |  0.6  |  0.0  |  0.4  
                cell3 |  0.0  |  0.1  |  0.0  |  0.0  
                =======================================

        labels_df(pandas.DataFrame): 
            sample * 2 matrix. 
            col_names must be (index, cell_type)
            Col1 is sample name the same as counts_df. 
            Col2 is celltypes or labels
            i.e.:
                =================
                sample | celltype
                =================
                cell1  |  B cell 
                cell2  |  T cell 
                cell3  |  B cell
                =================



        complex_table(pandas.DataFrame): 
            n * 2 matrix. 
            Col1 is complex compound ID.
            Col2 protein ID.
            i.e.:
                C1 is composed of gene1, gene2, gene3. The data is:
                ===========
                cid |  pid
                ===========
                C1  | gene1
                C1  | gene2
                C1  | gene3
                ===========


    output:
        mean_counts(pandas.DataFrame):
            celltype * feature(with complex feature) matrix. 
            item = average level of expression for each feature within each celltype. 
    '''
    ###################### Check Data ##########################



    ############################################################
    counts_df_with_labels = counts_df.merge(labels_df, left_index=True, right_index=True)
    mean_counts = counts_df_with_labels.groupby('cell_type', as_index=True, observed=True).mean()
    return mean_counts


def calculate_cluster_quantile(counts_df, labels_df, qt=0.9):
    '''
    input:
        counts_df(pandas.DataFrame): 
            sample * feature matrix. 
            The expression level of every gene in every cell.
            i.e.:
                =======================================
                      | gene1 | gene2 | gene3 | gene4 
                =======================================
                cell1 |  0.4  |  0.1  |  0.8  |  0.0
                cell2 |  0.0  |  0.6  |  0.0  |  0.4  
                cell3 |  0.0  |  0.1  |  0.0  |  0.0  
                =======================================

        labels_df(pandas.DataFrame): 
            sample * 2 matrix. 
            col_names must be (index, cell_type)
            Col1 is sample name the same as counts_df. 
            Col2 is celltypes or labels
            i.e.:
                =================
                sample | celltype
                =================
                cell1  |  B cell 
                cell2  |  T cell 
                cell3  |  B cell
                =================



        complex_table(pandas.DataFrame): 
            n * 2 matrix. 
            Col1 is complex compound ID.
            Col2 protein ID.
            i.e.:
                C1 is composed of gene1, gene2, gene3. The data is:
                ===========
                cid |  pid
                ===========
                C1  | gene1
                C1  | gene2
                C1  | gene3
                ===========


    output:
        mean_counts(pandas.DataFrame):
            celltype * feature(with complex feature) matrix. 
            item = average level of expression for each feature within each celltype. 
    '''

    ###################### Check Data ##########################



    ############################################################

    counts_df_with_labels = counts_df.merge(labels_df, left_index=True, right_index=True)
    # mean_counts = counts_df_with_labels.groupby('cell_type').quantile(qt)
    mean_counts = counts_df_with_labels.groupby('cell_type', observed=True).apply(lambda x: pd.Series(np.quantile(x,qt, axis=0, method='lower'), index=x.columns), include_groups=False)
    if 'cell_type' in mean_counts:
        mean_counts.drop(columns=['cell_type'], inplace=True)
    return mean_counts


def combine_complex_distribution_df(mean_counts, complex_table, complex_count_func):
    func = partial(complex_count_func, mean_counts=mean_counts)
    # complex count dataframe 
    if not complex_table.empty:
        complex_counts = complex_table.groupby('complex_multidata_id').apply(
            lambda x: x['protein_multidata_id'].values, include_groups=False).apply(
            lambda x:func(x)).T
        # 合成 最终 mean_counts dataframe
        mean_counts = pd.concat((mean_counts, complex_counts), axis=1)
    ##############################################################
    #                         Return                             #
    ##############################################################
    mean_counts = mean_counts.dropna(axis=1)
    return mean_counts

def calculate_complex_min_func(x, mean_counts):
    x = [sub_x for sub_x in x if sub_x in mean_counts.columns]
    if len(x) == 0:
        return pd.Series(index=mean_counts.index)
    return mean_counts.loc[:,x].T.min()

def calculate_complex_mean_func(x, mean_counts):
    x = [sub_x for sub_x in x if sub_x in mean_counts.columns]
    if len(x) == 0:
        return pd.Series(index=mean_counts.index)
    return mean_counts.loc[:,x].T.mean()


def calculate_interactions_strength(mean_counts, interactions, method='Arithmetic', sep='|'):
    assert method in ['Arithmetic', 'Geometric'], "Only support 'Arithmetic' or 'Geometric' mean."
    
    p1_index = []
    p2_index = []
    all_index = []
    for i in itertools.product(sorted(mean_counts.index), sorted(mean_counts.index)):
        p1_index.append(i[0])
        p2_index.append(i[1])
        all_index.append(sep.join(i))
        
    p1 = mean_counts.loc[p1_index, interactions['multidata_1_id']]
    p2 = mean_counts.loc[p2_index, interactions['multidata_2_id']]
    p1.columns = interactions.index
    p2.columns = interactions.index
    p1.index = all_index
    p2.index = all_index
    
    if method == 'Arithmetic':
        interactions_strength = (p1 + p2)/2 * (p1 > 0) * (p2>0)
    elif method == 'Geometric':
        interactions_strength = np.sqrt(p1*p2)
        
    return interactions_strength


import pandas as pd
import numpy as np
from .distrib import Distribution, get_pvalue_from_pmf
from collections import Counter
import itertools
import timeit
from .ccc_utils import get_current_memory, create_significant_interactions_df
from . import preprocess
from . import score
from . import dist_iid_set
from . import dist_complex 
from . import dist_lr
import warnings
import datetime
import os
from .cauchy_combine import cauthy_combine
import uuid
from loguru import logger
import glob

def generate_task_id(length=6):
    unique_part = uuid.uuid4().hex[:length]
    return unique_part

def mkdir(directory_path):
    if not os.path.exists(directory_path):
        os.makedirs(directory_path)
        logger.info(f"Directory created: {directory_path}")
    else:
        logger.info(f"Directory already exists: {directory_path}")


def __save(dataframe, save_path, task_id, timestamp, method_key, suffix, index=True):
    if method_key == '':
        save_file = os.path.join(save_path, f"{task_id}_{suffix}.tsv")
    else:
        save_file = os.path.join(save_path, f"{task_id}_{timestamp}_{method_key}_{suffix}.tsv")
    dataframe.to_csv(save_file, sep='\t', index=index)

def __save_file(
    interactions_strength, 
    pvals, 
    percents_analysis, 
    save_path, 
    task_id,
    timestamp = None,
    method_key = ''
):
    assert os.path.isdir(save_path), "{save_path} doesn't exist or not a dir"
    if timestamp is None:
        now = datetime.datetime.now()
        timestamp = now.strftime("%Y%m%d_%H%M%S")
    __save(interactions_strength, save_path, task_id, timestamp, method_key, 'interactions_strength')
    __save(pvals, save_path, task_id, timestamp, method_key, 'pvals')
    __save(percents_analysis, save_path, task_id, timestamp, method_key, 'percents_analysis')


def Cauchy_combination_of_statistical_analysis_methods(
    database_file_path,
    celltype_file_path,
    counts_file_path,
    convert_type = 'hgnc_symbol',
    single_unit_summary_list = ['Mean', 'Median', 'Q3', 'Quantile_0.9'],
    complex_aggregation_list = ['Minimum', 'Average'],
    LR_combination_list = ['Arithmetic', 'Geometric'],
    min_percentile = 0.1,
    save_path = None,
    meta_key = None, 
    select_list = [], 
    filter_ = False,
    use_DEG = False
):
    # Generate unique task ID.
    task_id = generate_task_id()
    logger.info(f"Task id is {task_id}.")

    # Set save directory.
    if save_path is None:
        save_path = './results/'
        logger.info(f"Results will be saved to \"{save_path}\".")
    mkdir(save_path)

    method_qt_dict = {
        "Median": 0.5,
        "Q2": 0.5,
        "Q3": 0.75,
    }

    counts_df, labels_df, complex_table, interactions = preprocess.get_input_data(
        database_file_path, 
        celltype_file_path,
        counts_file_path,
        convert_type,
        meta_key = meta_key,
        select_list = select_list,
        filter_ = filter_
    )
    logger.success("Data preprocessing done.")

    for cluster_distrib_method in single_unit_summary_list:
        cluster_distrib_key = cluster_distrib_method
        current_min_percentile = min_percentile
        # check input
        if cluster_distrib_method.startswith('Quantile_'):
            try:
                quantile = float(cluster_distrib_method.split('_')[-1])
                method_qt_dict["Quantile"] = quantile
                cluster_distrib_method = "Quantile"
            except ValueError:
                print('Quantile parameter in cluster_distrib_metho_list should be e.g. "Quantile_0.9"')
                raise ValueError
        if cluster_distrib_method in ["Median", "Q2", "Q3", "Quantile"]:
            quantile = method_qt_dict[cluster_distrib_method]
            cluster_distrib_method = 'Quantile'
            assert quantile < 1 and quantile > 0, "The quantile should be within the range (0, 1)."
            if quantile < 0.5:
                warnings.warn(f"Invalid quantile: {quantile}. The quantile is too low to be effective.", UserWarning)
            current_min_percentile = max(current_min_percentile, 1-quantile)
            logger.debug(f"Adjusted percentile is {current_min_percentile}")

        # Stage I : Percentages
        percents = calculate_cluster_percents(counts_df, labels_df, complex_table)
        percents_analysis = analyze_interactions_percents(percents, interactions, threshold=current_min_percentile)

        # Stage II : Counts Distribution
        ## Scores for clusters
        if cluster_distrib_method == 'Mean':
            mean_counts = score.calculate_cluster_mean(counts_df, labels_df)
        elif cluster_distrib_method == 'Quantile':
            mean_counts = score.calculate_cluster_quantile(counts_df, labels_df, quantile)

        # Stage III : Null Distribution
        if cluster_distrib_method == 'Mean':
            mean_pmfs = dist_iid_set.calculate_cluster_mean_distribution(counts_df, labels_df)
        elif cluster_distrib_method == 'Quantile':
            mean_pmfs = dist_iid_set.calculate_cluster_quantile_distribution(counts_df, labels_df, quantile)

        for complex_distrib_method in complex_aggregation_list:
            ## Counts for complex
            if complex_distrib_method == 'Minimum':
                complex_func = score.calculate_complex_min_func
            elif complex_distrib_method == 'Average':
                complex_func = score.calculate_complex_mean_func
            mean_counts_with_complex = score.combine_complex_distribution_df(mean_counts, complex_table, complex_func)
            ## Pmfs for complex
            if complex_distrib_method == 'Minimum':
                complex_func = dist_complex.get_minimum_distribution
            elif complex_distrib_method == 'Average':
                complex_func = dist_complex.get_average_distribution
            mean_pmfs_with_complex = dist_complex.combine_complex_distribution_df(mean_pmfs, complex_table, complex_func)
        

            for LR_distrib_method in LR_combination_list:
                method_key = f'{cluster_distrib_key}_{complex_distrib_method}_{LR_distrib_method}'

                logger.info(f"Running:\n-> {cluster_distrib_key} for single-unit summary function.\n"
                    + f"-> {complex_distrib_method} for multi-unit complex aggregation.\n"
                    + f"-> {LR_distrib_method} for L-R combination to compute the CS.\n"
                    + f"-> Percentile is {current_min_percentile}.")

                # Stage IV : P-values
                ## calculate L-R expression score:
                interactions_strength = score.calculate_interactions_strength(mean_counts_with_complex, interactions, method=LR_distrib_method)
                ## P-values for L-R expression:
                pvals = dist_lr.calculate_key_interactions_pvalue(
                    mean_pmfs_with_complex, interactions, interactions_strength, percents_analysis, method=LR_distrib_method
                )
                __save_file(interactions_strength, pvals, percents_analysis, save_path, task_id=task_id, method_key=method_key)
                logger.success("CS scoring module calculation done.")      
    logger.success("All scoring modules calculation done.")  
    cauthy_combine(save_path, task_id=task_id)
    logger.success("Cauthy combination done.") 

    pvals = pd.read_csv(os.path.join(save_path, f'{task_id}_Cauchy_pvals.tsv'), index_col=0, sep='\t')
    if use_DEG:
        mean_counts = score.calculate_cluster_mean(counts_df, labels_df)
        complex_func = score.calculate_complex_min_func
        mean_counts = score.combine_complex_distribution_df(mean_counts, complex_table, complex_func)
        mean_pmfs = dist_iid_set.calculate_cluster_mean_distribution(counts_df, labels_df)
        complex_func = dist_complex.get_minimum_distribution
        mean_pmfs = dist_complex.combine_complex_distribution_df(mean_pmfs, complex_table, complex_func)
        pvals = check_key_interactions_pvalue_by_DEG(mean_counts, mean_pmfs, interactions, pvals)
        pvals.to_csv(os.path.join(save_path, f'{task_id}_Cauchy_with_DEG_pvals.tsv'), sep='\t')
        logger.success("DEGs selection done.")
    
    significant_results = create_significant_interactions_df(pvals, database_file_path)
    __save(
        significant_results, 
        save_path = save_path,
        task_id = task_id, 
        method_key = '', 
        timestamp = '',
        suffix = 'significant_results',
        index = False
    )

    file_list = glob.glob(f'{save_path}/{task_id}*interactions_strength.tsv')
    logger.info(f"Integrating {len(file_list)} interactions_strength files.")
    average_strength_df = None
    for item in file_list:
        strength_df = pd.read_csv(item, index_col=0, sep='\t')
        if average_strength_df is None:
            average_strength_df = strength_df.copy()
        else:
            average_strength_df += strength_df.copy()
    average_strength_df = average_strength_df / len(file_list)
    average_strength_df.to_csv(os.path.join(save_path, f'{task_id}_average_interactions_strength.tsv'), sep='\t')
    logger.success("Average CS across all scoring methods calculation done.")






def statistical_analysis_method(
    database_file_path,
    celltype_file_path,
    counts_file_path,
    convert_type = 'hgnc_symbol',
    single_unit_summary = 'Mean',
    complex_aggregation = 'Minimum',
    LR_combination = 'Arithmetic',
    min_percentile = 0.1,
    style = None,
    meta_key=None, 
    select_list=[], 
    filter_= False,
    use_DEG = False,
    save_path = None
):
    # Generate unique task ID.
    task_id = generate_task_id()
    logger.info(f"Task id is {task_id}.")

    # Set save directory.
    if save_path is None:
        save_path = './results/'
        logger.info(f"Results will be saved to \"{save_path}\".")
    mkdir(save_path)

    method_qt_dict = {
        "Median": 0.5,
        "Q2": 0.5,
        "Q3": 0.75,
    }

    single_unit_key = single_unit_summary

    if single_unit_summary.startswith("Quantile_"):
        quantile = float(single_unit_summary.split("Quantile_")[1])
        single_unit_summary = "Quantile"

    if single_unit_summary in ["Median", "Q2", "Q3", "Quantile"]:
        if single_unit_summary in method_qt_dict:
            quantile = method_qt_dict[single_unit_summary]
            single_unit_summary = 'Quantile'
        assert quantile < 1 and quantile > 0, "The quantile should be within the range (0, 1)."
        if quantile < 0.5:
            warnings.warn(f"Invalid quantile: {quantile}. The quantile is too low to be effective.", UserWarning)
        min_percentile = max(min_percentile, 1-quantile)
    
    if style not in {None, "cpdb"}:
        raise ValueError(f"Invalid style: {style}. Must be one of [None, 'cpdb']")
    if style == 'cpdb':
        single_unit_summary = 'Mean'
        complex_aggregation = 'Minimum'
        LR_combination = 'Arithmetic'
        min_percentile = 0.1

    method_key = f'{single_unit_key}_{complex_aggregation}_{LR_combination}'
    
        
    counts_df, labels_df, complex_table, interactions = preprocess.get_input_data(
        database_file_path, 
        celltype_file_path,
        counts_file_path,
        convert_type,
        meta_key = meta_key,
        select_list = select_list,
        filter_ = filter_
    )
    logger.success("Data preprocessing done.")


    logger.info(f"Running:\n-> {single_unit_key} for single-unit summary function.\n"
                    + f"-> {complex_aggregation} for multi-unit complex aggregation.\n"
                    + f"-> {LR_combination} for L-R combination to compute the CS.\n"
                    + f"-> Percentile is {min_percentile}.")
    
    # Stage I : calculate L-R expression score:
    ## Scores for clusters
    if single_unit_summary == 'Mean':
        mean_counts = score.calculate_cluster_mean(counts_df, labels_df)
    elif single_unit_summary == 'Quantile':
        mean_counts = score.calculate_cluster_quantile(counts_df, labels_df, quantile)
    ## Scores for complex
    if complex_aggregation == 'Minimum':
        complex_func = score.calculate_complex_min_func
    elif complex_aggregation == 'Average':
        complex_func = score.calculate_complex_mean_func
    mean_counts = score.combine_complex_distribution_df(mean_counts, complex_table, complex_func)
    ## Scores for L-R expression
    interactions_strength = score.calculate_interactions_strength(mean_counts, interactions, method=LR_combination)
    
    # Stage II : Percentages
    percents = calculate_cluster_percents(counts_df, labels_df, complex_table)
    percents_analysis = analyze_interactions_percents(percents, interactions, threshold=min_percentile)

    # Stage III : Null Distribution
    if single_unit_summary == 'Mean':
        mean_pmfs = dist_iid_set.calculate_cluster_mean_distribution(counts_df, labels_df)
    elif single_unit_summary == 'Quantile':
        mean_pmfs = dist_iid_set.calculate_cluster_quantile_distribution(counts_df, labels_df, quantile)
    if complex_aggregation == 'Minimum':
        complex_func = dist_complex.get_minimum_distribution
    elif complex_aggregation == 'Average':
        complex_func = dist_complex.get_average_distribution
    mean_pmfs = dist_complex.combine_complex_distribution_df(mean_pmfs, complex_table, complex_func)
    
    # Stage IV : P-values
    pvals = dist_lr.calculate_key_interactions_pvalue(
        mean_pmfs, interactions, interactions_strength, percents_analysis, method=LR_combination
    )
    logger.success("FastCCC calculation done.")

    __save_file(interactions_strength, pvals, percents_analysis, save_path, task_id=task_id, method_key=method_key)

    if use_DEG:
        pvals = check_key_interactions_pvalue_by_DEG(mean_counts, mean_pmfs, interactions, pvals)
        pvals.to_csv(os.path.join(save_path, f'{task_id}_DEG_pvals.tsv'), sep='\t')
        logger.success("DEGs selection done.")

    significant_results = create_significant_interactions_df(pvals, database_file_path)
    __save(
        significant_results, 
        save_path = save_path,
        task_id = task_id, 
        method_key = '', 
        timestamp = '',
        suffix = 'significant_results',
        index = False
    )

    return interactions_strength, pvals, percents_analysis

# --- 06/27/2024 @marvinquiet
def cluster_markers_method(counts_file_path, celltype_file_path,
                           cluster_distrib_method = 'Mean',
                           quantile=0.9):
    '''Calculate cluster markers pvalues based on mean counts and null distribution
    input:
        counts_file_path(str): log-normalized count matrix path
        celltype_file_path(str): metadata file path that contains cell labels
        cluster_distrib_method(str): Mean/Quantile, method to calculate cluster distribution
        quantile(float): quantile to calculate the cluster distribution
    output:
        mean_counts(pandas.DataFrame): celltype * gene mean counts matrix.
        mean_pmfs(pandas.DataFrame): celltype * gene distribution matrix.
        clustermarker_pvals(pandas.DataFrame): celltype * gene marker pvalues matrix.
    '''
    counts_df, labels_df = preprocess.get_count_data(counts_file_path, celltype_file_path)
    # Preprocess input parameters
    method_qt_dict = {
        "Median": 0.5,
        "Q2": 0.5,
        "Q3": 0.75,
        "Quantile": quantile
    }
    if cluster_distrib_method in ["Median", "Q2", "Q3", "Quantile"]:
        quantile = method_qt_dict[cluster_distrib_method]
        cluster_distrib_method = 'Quantile'
        assert quantile < 1 and quantile > 0, "The quantile should be within the range (0, 1)."
        if quantile < 0.5:
            warnings.warn(f"Invalid quantile: {quantile}. The quantile is too low to be effective.", UserWarning)
    # Cluster Mean Counts
    if cluster_distrib_method == 'Mean':
        mean_counts = score.calculate_cluster_mean(counts_df, labels_df)
    elif cluster_distrib_method == 'Quantile':
        mean_counts = score.calculate_cluster_quantile(counts_df, labels_df, quantile)
    # Null Distribution
    if cluster_distrib_method == 'Mean':
        mean_pmfs = dist_iid_set.calculate_cluster_mean_distribution(counts_df, labels_df)
    elif cluster_distrib_method == 'Quantile':
        mean_pmfs = dist_iid_set.calculate_cluster_quantile_distribution(counts_df, labels_df, quantile)
    # get pvalues
    clustermarker_pvals = pd.DataFrame(
        list(map(get_pvalue_from_pmf, mean_counts.stack(), mean_pmfs.stack())),
        index=mean_counts.stack().index
    ).unstack()
    clustermarker_pvals = clustermarker_pvals.droplevel(0, axis=1) # drop the additional level generated by stack/unstack
    clustermarker_pvals = clustermarker_pvals.fillna(1.0) # fill NaN with 1.0
    return mean_counts, mean_pmfs, clustermarker_pvals
# --- end


def calculate_cluster_mean(counts_df, labels_df, complex_table):
    '''
    input:
        counts_df(pandas.DataFrame): 
            sample * feature matrix. 
            The expression level of every gene in every cell.
            i.e.:
                =======================================
                      | gene1 | gene2 | gene3 | gene4 
                =======================================
                cell1 |  0.4  |  0.1  |  0.8  |  0.0
                cell2 |  0.0  |  0.6  |  0.0  |  0.4  
                cell3 |  0.0  |  0.1  |  0.0  |  0.0  
                =======================================

        labels_df(pandas.DataFrame): 
            sample * 2 matrix. 
            col_names must be (index, cell_type)
            Col1 is sample name the same as counts_df. 
            Col2 is celltypes or labels
            i.e.:
                =================
                sample | celltype
                =================
                cell1  |  B cell 
                cell2  |  T cell 
                cell3  |  B cell
                =================



        complex_table(pandas.DataFrame): 
            n * 2 matrix. 
            Col1 is complex compound ID.
            Col2 protein ID.
            i.e.:
                C1 is composed of gene1, gene2, gene3. The data is:
                ===========
                cid |  pid
                ===========
                C1  | gene1
                C1  | gene2
                C1  | gene3
                ===========


    output:
        mean_counts(pandas.DataFrame):
            celltype * feature(with complex feature) matrix. 
            item = average level of expression for each feature within each celltype. 
    '''

    ###################### Check Data ##########################



    ############################################################

    counts_df_with_labels = counts_df.merge(labels_df, left_index=True, right_index=True)
    mean_counts = counts_df_with_labels.groupby('cell_type').mean()

    # complex count dataframe 
    def create_complex_count_func(x):
        x = [sub_x for sub_x in x if sub_x in mean_counts.columns]
        if len(x) == 0:
            return pd.Series(index=mean_counts.index)
        return mean_counts.loc[:,x].T.min()
    
    if not complex_table.empty:
        complex_counts = complex_table.groupby('complex_multidata_id').apply(
            lambda x: x['protein_multidata_id'].values, include_groups=False).apply(
            lambda x:create_complex_count_func(x)).T

        # 合成 最终 mean_counts dataframe
        mean_counts = pd.concat((mean_counts, complex_counts), axis=1)
    
    ##############################################################
    #                         Return                             #
    ##############################################################
    mean_counts = mean_counts.dropna(axis=1)
    return mean_counts


def calculate_cluster_quantile(counts_df, labels_df, complex_table, qt=0.9):
    '''
    input:
        counts_df(pandas.DataFrame): 
            sample * feature matrix. 
            The expression level of every gene in every cell.
            i.e.:
                =======================================
                      | gene1 | gene2 | gene3 | gene4 
                =======================================
                cell1 |  0.4  |  0.1  |  0.8  |  0.0
                cell2 |  0.0  |  0.6  |  0.0  |  0.4  
                cell3 |  0.0  |  0.1  |  0.0  |  0.0  
                =======================================

        labels_df(pandas.DataFrame): 
            sample * 2 matrix. 
            col_names must be (index, cell_type)
            Col1 is sample name the same as counts_df. 
            Col2 is celltypes or labels
            i.e.:
                =================
                sample | celltype
                =================
                cell1  |  B cell 
                cell2  |  T cell 
                cell3  |  B cell
                =================



        complex_table(pandas.DataFrame): 
            n * 2 matrix. 
            Col1 is complex compound ID.
            Col2 protein ID.
            i.e.:
                C1 is composed of gene1, gene2, gene3. The data is:
                ===========
                cid |  pid
                ===========
                C1  | gene1
                C1  | gene2
                C1  | gene3
                ===========


    output:
        mean_counts(pandas.DataFrame):
            celltype * feature(with complex feature) matrix. 
            item = average level of expression for each feature within each celltype. 
    '''

    ###################### Check Data ##########################



    ############################################################

    counts_df_with_labels = counts_df.merge(labels_df, left_index=True, right_index=True)
    # mean_counts = counts_df_with_labels.groupby('cell_type').quantile(qt)
    mean_counts = counts_df_with_labels.groupby('cell_type').apply(lambda x: pd.Series(np.quantile(x,qt, axis=0, method='lower'), index=x.columns), include_groups=False)


    # complex count dataframe 
    def create_complex_count_func(x):
        x = [sub_x for sub_x in x if sub_x in mean_counts.columns]
        if len(x) == 0:
            return pd.Series(index=mean_counts.index)
        return mean_counts.loc[:,x].T.min()
    
    if not complex_table.empty:
        complex_counts = complex_table.groupby('complex_multidata_id').apply(
            lambda x: x['protein_multidata_id'].values, include_groups=False).apply(
            lambda x:create_complex_count_func(x)).T

        # 合成 最终 mean_counts dataframe
        mean_counts = pd.concat((mean_counts, complex_counts), axis=1)
    
    ##############################################################
    #                         Return                             #
    ##############################################################
    mean_counts = mean_counts.dropna(axis=1)
    return mean_counts




def calculate_cluster_percents(counts_df, labels_df, complex_table):
    '''
    input:
        counts_df(pandas.DataFrame): 
            sample * feature matrix. 
            The expression level of every gene in every cell.
            i.e.:
                =======================================
                      | gene1 | gene2 | gene3 | gene4 
                =======================================
                cell1 |  0.4  |  0.1  |  0.8  |  0.0
                cell2 |  0.0  |  0.6  |  0.0  |  0.4  
                cell3 |  0.0  |  0.1  |  0.0  |  0.0  
                =======================================

        labels_df(pandas.DataFrame): 
            sample * 2 matrix. 
            col_names must be (index, cell_type)
            Col1 is sample name the same as counts_df. 
            Col2 is celltypes or labels
            i.e.:
                =================
                sample | celltype
                =================
                cell1  |  B cell 
                cell2  |  T cell 
                cell3  |  B cell
                =================



        complex_table(pandas.DataFrame): 
            n * 2 matrix. 
            Col1 is complex compound ID.
            Col2 protein ID.
            i.e.:
                C1 is composed of gene1, gene2, gene3. The data is:
                ===========
                cid |  pid
                ===========
                C1  | gene1
                C1  | gene2
                C1  | gene3
                ===========


    output:
        mean_counts(pandas.DataFrame):
            celltype * feature(with complex feature) matrix. 
            item = average level of expression for each feature within each celltype. 
    '''

    ###################### Check Data ##########################



    ############################################################
    
    counts_df = counts_df > 0
    counts_df_with_labels = counts_df.merge(labels_df, left_index=True, right_index=True)
    mean_counts = counts_df_with_labels.groupby('cell_type', as_index=True, observed=True).mean()

    # complex count dataframe 
    def create_complex_count_func(x):
        x = [sub_x for sub_x in x if sub_x in mean_counts.columns]
        if len(x) == 0:
            return pd.Series(index=mean_counts.index)
        return mean_counts.loc[:,x].T.min()
    if not complex_table.empty:
        complex_counts = complex_table.groupby('complex_multidata_id').apply(
            lambda x: x['protein_multidata_id'].values, include_groups=False).apply(
            lambda x:create_complex_count_func(x)).T

        # 合成 最终 mean_counts dataframe
        mean_counts = pd.concat((mean_counts, complex_counts), axis=1)
    
    ##############################################################
    #                         Return                             #
    ##############################################################
    mean_counts = mean_counts.dropna(axis=1)
    return mean_counts




# def calculate_cluster_mean_distribution(counts_df, labels_df, complex_table):
#     meta_dict = Counter(labels_df.cell_type)
    
#     ####### 
#     gene_sum_pmf_dict = {}
#     for gene in counts_df.columns:
#         samples = counts_df[gene].values
#         gene_sum_pmf_dict[gene] = {1: get_distribution_from_samples(samples)}
#         for item in range(2,30):
#             gene_sum_pmf_dict[gene][item] = gene_sum_pmf_dict[gene][item-1] + gene_sum_pmf_dict[gene][1]
            
#     for gene in counts_df.columns:
#         for item in range(2,30):
#             gene_sum_pmf_dict[gene][item] = gene_sum_pmf_dict[gene][item] / item
            
#     ####### clusters_mean #######
#     clusters_mean_dict = {}
#     for celltype in meta_dict:
#         clusters_mean_dict[celltype]  = {}
#         n_sum = meta_dict[celltype]
#         if n_sum < 30:
#             for gene in counts_df.columns:
#                 clusters_mean_dict[celltype][gene] = gene_sum_pmf_dict[gene][n_sum]
#         else:
#             for gene in counts_df.columns:
#                 clusters_mean_dict[celltype][gene] = gene_sum_pmf_dict[gene][1] ** n_sum / n_sum
#     mean_pmf = pd.DataFrame(clusters_mean_dict).T
    
#     def sub_func(x):
#         return get_minimum_distribution(*x)

#     def func(x):
#         x = [sub_x for sub_x in x if sub_x in mean_pmf.columns]
#         if len(x) == 0:
#             return pd.Series(index=mean_pmf.index)
#         return mean_pmf.loc[:,x].apply(sub_func, axis=1)
    
#     if not complex_table.empty:
#         complex_pmf = complex_table.groupby('complex_multidata_id').apply(lambda x: x['protein_multidata_id'].values).apply(lambda x:func(x)).T
#         complex_pmf = complex_pmf.dropna(axis=1)
#         mean_pmf = pd.concat((mean_pmf, complex_pmf), axis=1)
        
#     return mean_pmf


def calculate_interactions_strength(mean_counts, interactions, sep='|'):
    
    # 
    p1_index = []
    p2_index = []
    all_index = []
    for i in itertools.product(sorted(mean_counts.index), sorted(mean_counts.index)):
        p1_index.append(i[0])
        p2_index.append(i[1])
        all_index.append(sep.join(i))
        
    p1 = mean_counts.loc[p1_index, interactions['multidata_1_id']]
    p2 = mean_counts.loc[p2_index, interactions['multidata_2_id']]
    p1.columns = interactions.index
    p2.columns = interactions.index
    p1.index = all_index
    p2.index = all_index
    
    interactions_strength = (p1 + p2)/2 * (p1 > 0) * (p2>0)
    
    return interactions_strength

def calculate_interactions_strength_multiply(mean_counts, interactions, sep='|'):
    
    # 
    p1_index = []
    p2_index = []
    all_index = []
    for i in itertools.product(sorted(mean_counts.index), sorted(mean_counts.index)):
        p1_index.append(i[0])
        p2_index.append(i[1])
        all_index.append(sep.join(i))
        
    p1 = mean_counts.loc[p1_index, interactions['multidata_1_id']]
    p2 = mean_counts.loc[p2_index, interactions['multidata_2_id']]
    p1.columns = interactions.index
    p2.columns = interactions.index
    p1.index = all_index
    p2.index = all_index
    
    interactions_strength = np.sqrt(p1*p2)#(p1 + p2)/2 * (p1 > 0) * (p2>0)
    
    return interactions_strength


def analyze_interactions_percents(cluster_percents, interactions, threshold=0.1, sep='|'):
    
    # 
    p1_index = []
    p2_index = []
    all_index = []
    for i in itertools.product(sorted(cluster_percents.index), sorted(cluster_percents.index)):
        p1_index.append(i[0])
        p2_index.append(i[1])
        all_index.append(sep.join(i))
        
    p1 = cluster_percents.loc[p1_index, interactions['multidata_1_id']]
    p2 = cluster_percents.loc[p2_index, interactions['multidata_2_id']]
    p1.columns = interactions.index
    p2.columns = interactions.index
    p1.index = all_index
    p2.index = all_index
    
    interactions_strength = (p1>threshold) * (p2>threshold)
    # print((p1>threshold) * (p2>threshold))
    
    return interactions_strength


    
    
def calculate_key_interactions_pvalue(mean_pmf, interactions, interactions_strength, percent_analysis):
    p1_index = []
    p2_index = []
    all_index = []
    for i in itertools.product(sorted(mean_pmf.index), sorted(mean_pmf.index)):
        p1_index.append(i[0])
        p2_index.append(i[1])
        all_index.append('|'.join(i))
        
    p1 = mean_pmf.loc[p1_index, interactions['multidata_1_id']]
    p2 = mean_pmf.loc[p2_index, interactions['multidata_2_id']]
    p1.columns = interactions.index
    p2.columns = interactions.index
    p1.index = all_index
    p2.index = all_index
    
    p1_items = p1.values[np.where(percent_analysis)]
    p2_items = p2.values[np.where(percent_analysis)]
    
    
    start = timeit.default_timer()
    pval_pmfs = (p1_items & p2_items) / 2
    stop = timeit.default_timer()
    
    mean_gt = interactions_strength.values[np.where(percent_analysis)]
    est = []
    for i, value in enumerate(mean_gt):
        pval_est = get_pvalue_from_pmf(value, pval_pmfs[i])
        est.append(pval_est)
        
    pvalues = np.ones_like(interactions_strength)
    pvalues[np.where(percent_analysis)] = est
    pvalues = pd.DataFrame(pvalues, index=interactions_strength.index, columns=interactions_strength.columns)
    return pvalues


########################
def calculate_key_interactions_pvalue_multiply_version(mean_pmf, interactions, interactions_strength, percent_analysis):
    p1_index = []
    p2_index = []
    all_index = []
    for i in itertools.product(sorted(mean_pmf.index), sorted(mean_pmf.index)):
        p1_index.append(i[0])
        p2_index.append(i[1])
        all_index.append('|'.join(i))
        
    p1 = mean_pmf.loc[p1_index, interactions['multidata_1_id']]
    p2 = mean_pmf.loc[p2_index, interactions['multidata_2_id']]
    p1.columns = interactions.index
    p2.columns = interactions.index
    p1.index = all_index
    p2.index = all_index
    
    p1_items = p1.values[np.where(percent_analysis)]
    p2_items = p2.values[np.where(percent_analysis)]
    
    
    start = timeit.default_timer()
    pval_pmfs = p1_items * p2_items
    stop = timeit.default_timer()
    
    mean_gt = interactions_strength.values[np.where(percent_analysis)]
    est = []
    for i, value in enumerate(mean_gt):
        pval_est = get_pvalue_from_pmf(value, pval_pmfs[i])
        est.append(pval_est)
        
    pvalues = np.ones_like(interactions_strength)
    pvalues[np.where(percent_analysis)] = est
    pvalues = pd.DataFrame(pvalues, index=interactions_strength.index, columns=interactions_strength.columns)
    return pvalues
########################



def calculate_key_interactions_pvalue_by_part(mean_counts, mean_pmf, interactions, interactions_strength, percent_analysis):
    p1_index = []
    p2_index = []
    all_index = []
    for i in itertools.product(sorted(mean_pmf.index), sorted(mean_pmf.index)):
        p1_index.append(i[0])
        p2_index.append(i[1])
        all_index.append('|'.join(i))
        
    p1 = mean_pmf.loc[p1_index, interactions['multidata_1_id']]
    p2 = mean_pmf.loc[p2_index, interactions['multidata_2_id']]
    p1.columns = interactions.index
    p2.columns = interactions.index
    p1.index = all_index
    p2.index = all_index
    
    p1_items = p1.values[np.where(percent_analysis)]
    p2_items = p2.values[np.where(percent_analysis)]
    
    mean_count_p1 = mean_counts.loc[p1_index, interactions['multidata_1_id']]
    mean_count_p2 = mean_counts.loc[p2_index, interactions['multidata_2_id']]
    mean_count_p1.columns = interactions.index
    mean_count_p2.columns = interactions.index
    mean_count_p1.index = all_index
    mean_count_p2.index = all_index
    mean_count_p1_item = mean_count_p1.values[np.where(percent_analysis)]
    mean_count_p2_item = mean_count_p2.values[np.where(percent_analysis)]
    
    
    est1 = []
    for i, value in enumerate(mean_count_p1_item):
        pval_est = get_pvalue_from_pmf(value, p1_items[i])
        est1.append(pval_est)
    est1 = np.array(est1) < 0.05
        
    est2 = []
    for i, value in enumerate(mean_count_p2_item):
        pval_est = get_pvalue_from_pmf(value, p2_items[i])
        est2.append(pval_est)
    est2 = np.array(est2) < 0.05
        
    significant_flag_mat = np.zeros_like(interactions_strength, dtype=bool)
    significant_flag_mat[np.where(percent_analysis)] = np.logical_and(est1, est2)
    significant_flag_mat = pd.DataFrame(significant_flag_mat, index=interactions_strength.index, columns=interactions_strength.columns)
    return significant_flag_mat


def check_key_interactions_pvalue_by_DEG(mean_counts, mean_pmf, interactions, pvals):
    p1_index = []
    p2_index = []
    all_index = []
    for i in itertools.product(sorted(mean_pmf.index), sorted(mean_pmf.index)):
        p1_index.append(i[0])
        p2_index.append(i[1])
        all_index.append('|'.join(i))
        
    p1 = mean_pmf.loc[p1_index, interactions['multidata_1_id']]
    p2 = mean_pmf.loc[p2_index, interactions['multidata_2_id']]
    p1.columns = interactions.index
    p2.columns = interactions.index
    p1.index = all_index
    p2.index = all_index
    
    p1_items = p1.values[np.where(pvals.values < 0.05)]
    p2_items = p2.values[np.where(pvals.values < 0.05)]
    
    mean_count_p1 = mean_counts.loc[p1_index, interactions['multidata_1_id']]
    mean_count_p2 = mean_counts.loc[p2_index, interactions['multidata_2_id']]
    mean_count_p1.columns = interactions.index
    mean_count_p2.columns = interactions.index
    mean_count_p1.index = all_index
    mean_count_p2.index = all_index
    mean_count_p1_item = mean_count_p1.values[np.where(pvals.values < 0.05)]
    mean_count_p2_item = mean_count_p2.values[np.where(pvals.values < 0.05)]
    
    
    est1 = []
    for i, value in enumerate(mean_count_p1_item):
        pval_est = get_pvalue_from_pmf(value, p1_items[i])
        est1.append(pval_est)
    est1 = np.array(est1) < 0.05
        
    est2 = []
    for i, value in enumerate(mean_count_p2_item):
        pval_est = get_pvalue_from_pmf(value, p2_items[i])
        est2.append(pval_est)
    est2 = np.array(est2) < 0.05
        
    mask = np.float64(np.logical_and(est1, est2))
    masked_values = mask * pvals.values[np.where(pvals.values < 0.05)] + (1 - mask)
    pvals.values[np.where(pvals.values < 0.05)] = masked_values
    return pvals



import pandas as pd
import numpy as np
import anndata
import scanpy as sc
import os
from . import preproc_utils
from .ccc_utils import get_current_memory
import timeit
from loguru import logger


def get_interactions(cpdb_file_path, select_list=[]):
    interactions = pd.read_csv(os.path.join(cpdb_file_path, 'interaction_table.csv'))[['id_cp_interaction', 'multidata_1_id', 'multidata_2_id']]
    interactions.set_index('id_cp_interaction', inplace=True)
    interactions = interactions[['multidata_1_id', 'multidata_2_id']].drop_duplicates()
    if select_list:
        interactions = interactions.loc[select_list]
    return interactions


def get_input_data(cpdb_file_path, meta_file_path, counts_file_path, convert_type, meta_key=None, select_list=[], filter_=False):

    '''
    input:
        1. cpdb_file_path(str):
            a path of dir which contains all interaction data
            details see temp
        2. coutns_file_path(str):


    temp:
        I1 -> cpdb_file_path/interaction_table.csv -> interactions(pandas.DataFrame):
        =======================================================
        id_cp_interaction  |  multidata_1_id  |  multidata_2_id
        =======================================================
        CPI-CS0A5B6BD7A    |       1364       |       797
        CPI-CS047D9C0D7    |       1532       |       797
        CPI-CS04A56D5BE    |       1364       |       1004
        CPI-CS0F5B070C5    |       1532       |       1004
        =======================================================

        I2 -> counts(adata):
        main_frame is:
        ===============================
              | gene1 | gene2 | gene3
        ===============================
        cell1 |  0.5  |  0.6  |  0.0
        cell2 |  0.0  |  0.2  |  0.0   
        cell3 |  0.5  |  0.0  |  0.4   
        cell4 |  0.0  |  0.0  |  0.1  
        ===============================


    output:

    '''
    


    ######################################################################
    #                          1.Data Input                              #
    ######################################################################

    ### interactions, counts, labels  ##
    interactions = get_interactions(cpdb_file_path, select_list)
    start = timeit.default_timer()
    counts = anndata.read_h5ad(counts_file_path)#.to_df()
    counts.var_names_make_unique()
    stop = timeit.default_timer()
    logger.debug(f'Read Time: {stop - start}') 
    
    #MMMMM
    current_memory = get_current_memory()
    logger.debug("reading_count: {:.2f}MB".format(current_memory))
    #MMMMM
    # 0815 add:
    if filter_:
        sc.pp.filter_cells(counts, min_counts=1)
        sc.pp.filter_genes(counts, min_cells=1)

    
    if meta_key is not None:
        labels_df = pd.DataFrame(counts.obs[meta_key])
        labels_df.columns = ['cell_type']
        labels_df.index.name = 'barcode_sample'
    else:
        labels_df = pd.read_csv(meta_file_path, sep='\t', index_col=0)
    
    '''
    interactions(pandas.DataFrame):
    =======================================================
                      |  multidata_1_id  |  multidata_2_id
    id_cp_interaction |                  |            
    =======================================================
    CPI-CS0A5B6BD7A   |        1364      |       797
    CPI-CS047D9C0D7   |        1532      |       797
    CPI-CS04A56D5BE   |        1364      |      1004
    =======================================================
    '''
    ####################################



    ##### gene_table ########
    gene_table = pd.read_csv(os.path.join(cpdb_file_path, 'gene_table.csv'))
    protein_table = pd.read_csv(os.path.join(cpdb_file_path, 'protein_table.csv'))
    gene_table = gene_table.merge(protein_table, left_on='protein_id', right_on='id_protein')
    #########################


    ##### complex_table ######
    complex_composition = pd.read_csv(os.path.join(cpdb_file_path, 'complex_composition_table.csv'))
    complex_table = pd.read_csv(os.path.join(cpdb_file_path, 'complex_table.csv'))
    complex_table = complex_table.merge(complex_composition, left_on='complex_multidata_id', right_on='complex_multidata_id')

    # 让我们只关注 'complex_multidata_id'，'protein_multidata_id'
    complex_table = complex_table[['complex_multidata_id','protein_multidata_id']]
    '''
    complex_table(pandas.DataFrame):
    =======================================================
           | complex_multidata_id   |  protein_multidata_id
    -------------------------------------------------------
    0      |             1355       |          1134
    1      |             1356       |          1175
    2      |             1357       |          1167
    =======================================================
    '''
    ##########################


   


    ######################################################################
    #                      2.Counts DF preprocess                        #
    ######################################################################
    
    ##### feature to id conversion  ######
    # 不在 标准列表 里的 gene 就不要了
    tmp = gene_table[[convert_type, 'protein_multidata_id']]
    tmp = tmp.drop_duplicates()
    tmp.set_index('protein_multidata_id', inplace=True)

    select_columns = []
    columns_names = []
    for foo, boo in zip(tmp.index, tmp[convert_type]):
        if boo in counts.var_names:#counts.columns:
            select_columns.append(boo)
            columns_names.append(foo)

    reduced_counts = counts[:, select_columns].to_df()
    reduced_counts.columns = columns_names
    reduced_counts = reduced_counts.T.groupby(level=0).mean().T
    # FutureWarning: DataFrame.groupby with axis=1 is deprecated. Do `frame.T.groupby(...)` without axis instead.
    # reduced_counts = reduced_counts.groupby(reduced_counts.columns, axis=1).mean()
    ######################################
    
    
    ########## filter genes  ############
    # gene 在 所有 cell 上为 0 不要
    reduced_counts = preproc_utils.filter_empty_genes(reduced_counts)
    ######################################
    
    
    
    ######################################################################
    #                      3.Other DF filtered                       #
    ######################################################################
    # 一个 interaction 可能只有 partA 存在，但是 partB 不存在
    # 只有 如果 任意一部分不存在， 另一部分没必要参与后续计算
    
    ##### delete item not involved interactions ####
    
    foo_dict = complex_table.groupby('complex_multidata_id').apply(lambda x: list(x['protein_multidata_id'].values), include_groups=False).to_dict()
    '''
    dictionary complex_id: [protein_id_1, pid2, pid3, ...]
    foo_dict = {
        1355: [1134],
        1356: [1175],
        xxxx: [AAAA, BBBB, CCCC],
    }
    '''

    def __content__(key):
        if key not in foo_dict:
            return [key]
        else:
            return foo_dict[key]

    # def __exist__(key, df):
    #     # 目前的 complex 策略就是 有一个 就可以了
    #     for item in __content__(key):
    #         if item in df.columns:
    #             return True
    #     return False
    
    def __exist__(key, df):
        # 目前的 complex 策略就是 全部都要有
        # 经测试，这是cpdb用的策略
        for item in __content__(key):
            if item not in df.columns:
                return False
        return True
    
    # ###### 检验 __exist__ ####
    # select_index = []
    # for item in complex_table['complex_multidata_id']:
    #     if __exist__(item, reduced_counts):
    #         select_index.append(True)
    #     else:
    #         select_index.append(False)
    # complex_composition_filtered = complex_table[select_index]
    # #########################

    temp_list = []
    temp_dict = {}
    for item in reduced_counts.columns:
        temp_dict[item] = False
    
    # 注释是为了验证 interactions 的过滤策略， 完全一致
    # print(interactions)
    select_index = []
    for partA, partB in zip(interactions.multidata_1_id, interactions.multidata_2_id):
        if __exist__(partA, reduced_counts) and __exist__(partB, reduced_counts):
            temp_list.extend([partA, partB])
            select_index.append(True)
        else:
            select_index.append(False)
    interactions_filtered = interactions[select_index]
    
    ###### cpdb verification #######
#     select_cols = []
#     for item in set(temp_list):
#         if item not in foo_dict:
#             select_cols.append(item)
#     counts_simple = reduced_counts[select_cols]

#     select_cols = []
#     for item in set(temp_list):
#         if item in foo_dict:
#             select_cols.extend(__content__(item))
#     # 从这里开始出现了不一致， cpdb先看 complex是否存在 （所有protein组成部分都在），
#     # 然后生成了 counts_complex 部分，然后 和 simplex 拼接，
#     # 必然 出现多了一些没用的部分 
#     counts_complex = reduced_counts[list(set(select_cols))]
    ################################
    
            
    for item in temp_list:
        for subitem in __content__(item):
            if subitem in temp_dict:
                temp_dict[subitem] = True
    select_index = [key for key in temp_dict if temp_dict[key]]
    reduced_counts = reduced_counts[select_index]
    ################################################
    
    #MMMMM
    current_memory = get_current_memory()
    logger.debug("get_input_data_peak: {:.2f}MB".format(current_memory))
    #MMMMM
    
    ######################################################################
    #                               Return                               #
    ######################################################################
    counts_df = reduced_counts
    temp_list = set(temp_list)
    select_index = [True if item in temp_list else False for item in complex_table.complex_multidata_id]
    complex_table = complex_table[select_index]
    interactions = interactions_filtered
    return counts_df, labels_df, complex_table, interactions
    
    
    
    